(*
(C) Copyright Andreas Viktor Hess, DTU, 2018
(C) Copyright Sebastian A. Mödersheim, DTU, 2018
(C) Copyright Achim D. Brucker, University of Sheffield, 2018

All Rights Reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

- Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

- Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

- Neither the name of the copyright holder nor the names of its
  contributors may be used to endorse or promote products
  derived from this software without specific prior written
  permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*)

(*  Title:      ParallelCompositionality.thy
    Author:     Andreas Viktor Hess, DTU
    Author:     Sebastian A. Mödersheim, DTU
    Author:     Achim D. Brucker, The University of Sheffield
*)

theory ParallelCompositionality
imports ProtocolTransitionSystem "~~/src/HOL/Library/Sublist"
begin

section {* Parallel compositionality of security protocols *}

subsection {* Definitions: Labeled strands/constraints *}

lemmas prefixeq_append = prefixeq_prefixeq append_prefixeqD

definition (in typed_model) typing_cond where
  "typing_cond \<A> \<equiv>
    wf\<^sub>s\<^sub>t {} \<A> \<and>
    FV\<^sub>s\<^sub>t \<A> \<inter> bvars\<^sub>s\<^sub>t \<A> = {} \<and>
    tfr\<^sub>s\<^sub>t (trms\<^sub>s\<^sub>t \<A>) \<and> eqs_tfr\<^sub>s\<^sub>t \<A> \<and>
    Ana_invar_subst (ik\<^sub>s\<^sub>t \<A> \<union> eqs_rhs\<^sub>s\<^sub>t \<A>)"

datatype strand_label =
  Label1 ("l1")
| Label2 ("l2")
| LabelS ("\<star>")

type_synonym ('a,'b) labeledstrand_step = "strand_label \<times> ('a,'b) strand_step"
type_synonym ('a,'b) labeledstrand = "('a,'b) labeledstrand_step list"

abbreviation is_Label1 where "is_Label1 x \<equiv> \<exists>s. x = (l1, s)"
abbreviation is_Label2 where "is_Label2 x \<equiv> \<exists>s. x = (l2, s)"
abbreviation is_LabelS where "is_LabelS x \<equiv> \<exists>s. x = (\<star>, s)"

abbreviation labeledstrand_step_prepend (infixr ":" 65) where
  "(b::('f,'v) labeledstrand_step):A \<equiv> b#A"

abbreviation labeledstrand_step_append (infixr ":" 65) where 
  "A:(b::('f,'v) labeledstrand_step) \<equiv> A@[b]"

abbreviation labeledstrand_append (infixr ":" 65) where
  "(A::('f,'v) labeledstrand):B \<equiv> A@B"

abbreviation subst_apply_labeledstrand_step (infix "\<cdot>\<^sub>l\<^sub>s\<^sub>t\<^sub>p" 51) where
  "x \<cdot>\<^sub>l\<^sub>s\<^sub>t\<^sub>p \<theta> \<equiv> (case x of (l, s) \<Rightarrow> (l, s \<cdot>\<^sub>s\<^sub>t\<^sub>p \<theta>))"

abbreviation subst_apply_labeledstrand (infix "\<cdot>\<^sub>l\<^sub>s\<^sub>t" 51) where
  "S \<cdot>\<^sub>l\<^sub>s\<^sub>t \<theta> \<equiv> map (\<lambda>x. x \<cdot>\<^sub>l\<^sub>s\<^sub>t\<^sub>p \<theta>) S"

definition unlabel where "unlabel S \<equiv> map snd S"
definition left where "left S \<equiv> filter (\<lambda>s. \<not>is_Label2 s) S"
definition right where "right S \<equiv> filter (\<lambda>s. \<not>is_Label1 s) S"
abbreviation left_unl where "left_unl S \<equiv> unlabel (left S)"
abbreviation right_unl where "right_unl S \<equiv> unlabel (right S)"

abbreviation trms\<^sub>l\<^sub>s\<^sub>t where "trms\<^sub>l\<^sub>s\<^sub>t S \<equiv> trms\<^sub>s\<^sub>t (unlabel S)"
abbreviation trms_left\<^sub>l\<^sub>s\<^sub>t where "trms_left\<^sub>l\<^sub>s\<^sub>t S \<equiv> trms\<^sub>s\<^sub>t (left_unl S)"
abbreviation trms_right\<^sub>l\<^sub>s\<^sub>t where "trms_right\<^sub>l\<^sub>s\<^sub>t S \<equiv> trms\<^sub>s\<^sub>t (right_unl S)"

abbreviation vars\<^sub>l\<^sub>s\<^sub>t where "vars\<^sub>l\<^sub>s\<^sub>t S \<equiv> vars\<^sub>s\<^sub>t (unlabel S)"
abbreviation vars_left\<^sub>l\<^sub>s\<^sub>t where "vars_left\<^sub>l\<^sub>s\<^sub>t S \<equiv> vars\<^sub>s\<^sub>t (left_unl S)"
abbreviation vars_right\<^sub>l\<^sub>s\<^sub>t where "vars_right\<^sub>l\<^sub>s\<^sub>t S \<equiv> vars\<^sub>s\<^sub>t (right_unl S)"

abbreviation vars2\<^sub>l\<^sub>s\<^sub>t where "vars2\<^sub>l\<^sub>s\<^sub>t S \<equiv> vars2\<^sub>s\<^sub>t (unlabel S)"
abbreviation bvars\<^sub>l\<^sub>s\<^sub>t where "bvars\<^sub>l\<^sub>s\<^sub>t S \<equiv> bvars\<^sub>s\<^sub>t (unlabel S)"
abbreviation FV\<^sub>l\<^sub>s\<^sub>t where "FV\<^sub>l\<^sub>s\<^sub>t S \<equiv> FV\<^sub>s\<^sub>t (unlabel S)"

context typed_model
begin

subsection {* Definitions: Ground SMP and parallel composability *}

abbreviation GSMP\<^sub>s\<^sub>t::"('fun,'var) terms \<Rightarrow> ('fun,'var) term set" where
  "GSMP\<^sub>s\<^sub>t P \<equiv> {t \<in> SMP\<^sub>s\<^sub>t P. FV t = {}}"

definition GSMP_disjoint where
  "GSMP_disjoint P1 P2 Secrets \<equiv> GSMP\<^sub>s\<^sub>t P1 \<inter> GSMP\<^sub>s\<^sub>t P2 \<subseteq> Secrets \<union> {m. {} \<turnstile>\<^sub>c m}"

definition par_comp where
  "par_comp (\<A>::('fun,'var) labeledstrand) (Secrets::('fun,'var) terms) \<equiv> 
    GSMP_disjoint (trms_left\<^sub>l\<^sub>s\<^sub>t \<A>) (trms_right\<^sub>l\<^sub>s\<^sub>t \<A>) Secrets \<and>
    (\<forall>t. (\<star>, snd\<langle>t\<rangle>) \<notin> set \<A>) \<and> ground Secrets \<and>
    (\<forall>s \<in> Secrets. \<forall>s' \<in> subterms s. {} \<turnstile>\<^sub>c s' \<or> s' \<in> Secrets)"

definition declassified\<^sub>l\<^sub>s\<^sub>t where
  "declassified\<^sub>l\<^sub>s\<^sub>t (\<A>::('fun,'var) labeledstrand) \<I> \<equiv> {t. (\<star>, rcv\<langle>t\<rangle>) \<in> set \<A>} \<cdot>set \<I>"

definition strand_leaks\<^sub>l\<^sub>s\<^sub>t ("_ leaks _ under _") where
  "\<A> leaks Secrets under \<I> \<equiv>
    (\<exists>t \<in> Secrets - declassified\<^sub>l\<^sub>s\<^sub>t \<A> \<I>. (\<I> \<Turnstile> \<langle>left_unl \<A>:snd\<langle>t\<rangle>\<rangle>) \<or> (\<I> \<Turnstile> \<langle>right_unl \<A>:snd\<langle>t\<rangle>\<rangle>))"

definition strand_leaks_c\<^sub>l\<^sub>s\<^sub>t where
  "strand_leaks_c\<^sub>l\<^sub>s\<^sub>t \<A> Secrets \<I> \<equiv>
    (\<exists>t \<in> Secrets - declassified\<^sub>l\<^sub>s\<^sub>t \<A> \<I>. (\<I> \<Turnstile>\<^sub>c \<langle>left_unl \<A>:snd\<langle>t\<rangle>\<rangle>) \<or> (\<I> \<Turnstile>\<^sub>c \<langle>right_unl \<A>:snd\<langle>t\<rangle>\<rangle>))"


subsection {* Definitions: Homogeneous and numbered intruder deduction variants *}

definition left_specific where
  "left_specific t \<A> Secrets \<equiv> t \<in> GSMP\<^sub>s\<^sub>t (trms_left\<^sub>l\<^sub>s\<^sub>t \<A>) - (Secrets \<union> {m. {} \<turnstile>\<^sub>c m})"

definition right_specific where
  "right_specific t \<A> Secrets \<equiv> t \<in> GSMP\<^sub>s\<^sub>t (trms_right\<^sub>l\<^sub>s\<^sub>t \<A>) - (Secrets \<union> {m. {} \<turnstile>\<^sub>c m})"

definition heterogeneous\<^sub>l\<^sub>s\<^sub>t where
  "heterogeneous\<^sub>l\<^sub>s\<^sub>t t \<A> Secrets \<equiv> (
    (\<exists>s \<in> subterms t. left_specific  s \<A> Secrets) \<and>
    (\<exists>s \<in> subterms t. right_specific s \<A> Secrets))"

abbreviation homogeneous\<^sub>l\<^sub>s\<^sub>t where
  "homogeneous\<^sub>l\<^sub>s\<^sub>t t \<A> Secrets \<equiv> \<not>heterogeneous\<^sub>l\<^sub>s\<^sub>t t \<A> Secrets"

declare GSMP_disjoint_def[simp]
declare par_comp_def[simp]
declare declassified\<^sub>l\<^sub>s\<^sub>t_def[simp]
declare left_specific_def[simp]
declare right_specific_def[simp]
declare heterogeneous\<^sub>l\<^sub>s\<^sub>t_def[simp]

inductive intruder_deduct_hom::
  "('fun,'var) terms \<Rightarrow> ('fun,'var) labeledstrand \<Rightarrow> ('fun,'var) terms \<Rightarrow> ('fun,'var) term \<Rightarrow> bool"
  ("\<langle>_;_;_\<rangle> \<turnstile>\<^sub>h\<^sub>o\<^sub>m _" 50)
where
  AxiomH[simp]:   "t \<in> M \<Longrightarrow> \<langle>M; \<A>; Secrets\<rangle> \<turnstile>\<^sub>h\<^sub>o\<^sub>m t"
| ComposeH[simp]: "\<lbrakk>length X = arity f; public f; \<And>x. x \<in> set X \<Longrightarrow> \<langle>M; \<A>; Secrets\<rangle> \<turnstile>\<^sub>h\<^sub>o\<^sub>m x;
                    homogeneous\<^sub>l\<^sub>s\<^sub>t (Fun f X) \<A> Secrets;
                    Fun f X \<in> GSMP\<^sub>s\<^sub>t (trms\<^sub>l\<^sub>s\<^sub>t \<A>)\<rbrakk>
                    \<Longrightarrow> \<langle>M; \<A>; Secrets\<rangle> \<turnstile>\<^sub>h\<^sub>o\<^sub>m Fun f X"
| DecomposeH:     "\<lbrakk>\<langle>M; \<A>; Secrets\<rangle> \<turnstile>\<^sub>h\<^sub>o\<^sub>m t; Ana t = (K, T);
                    \<And>k. k \<in> K \<Longrightarrow> \<langle>M; \<A>; Secrets\<rangle> \<turnstile>\<^sub>h\<^sub>o\<^sub>m k; t\<^sub>i \<in> T\<rbrakk>
                    \<Longrightarrow> \<langle>M; \<A>; Secrets\<rangle> \<turnstile>\<^sub>h\<^sub>o\<^sub>m t\<^sub>i"

inductive intruder_deduct_num::
  "('fun,'var) terms \<Rightarrow> nat \<Rightarrow> ('fun,'var) term \<Rightarrow> bool"
  ("\<langle>_; _\<rangle> \<turnstile>\<^sub>n _" 50)
where
  AxiomN[simp]:   "t \<in> M \<Longrightarrow> \<langle>M; 0\<rangle> \<turnstile>\<^sub>n t"
| ComposeN[simp]: "\<lbrakk>length T = arity f; public f; \<And>t. t \<in> set T \<Longrightarrow> \<langle>M; steps t\<rangle> \<turnstile>\<^sub>n t\<rbrakk>
                    \<Longrightarrow> \<langle>M; Suc (Max (insert 0 (steps ` set T)))\<rangle> \<turnstile>\<^sub>n Fun f T"
| DecomposeN:     "\<lbrakk>\<langle>M; n\<rangle> \<turnstile>\<^sub>n t; Ana t = (K, T); \<And>k. k \<in> K \<Longrightarrow> \<langle>M; steps k\<rangle> \<turnstile>\<^sub>n k; t\<^sub>i \<in> T\<rbrakk>
                    \<Longrightarrow> \<langle>M; Suc (Max (insert n (steps ` K)))\<rangle> \<turnstile>\<^sub>n t\<^sub>i"

lemma intruder_deduct_hom_induct[consumes 1, case_names AxiomH ComposeH DecomposeH]:
  assumes "\<langle>M; \<A>; Sec\<rangle> \<turnstile>\<^sub>h\<^sub>o\<^sub>m t" "\<And>t. t \<in> M \<Longrightarrow> P M t"
          "\<And>X f. \<lbrakk>length X = arity f; public f;
                  \<And>x. x \<in> set X \<Longrightarrow> \<langle>M; \<A>; Sec\<rangle> \<turnstile>\<^sub>h\<^sub>o\<^sub>m x;
                  \<And>x. x \<in> set X \<Longrightarrow> P M x;
                  homogeneous\<^sub>l\<^sub>s\<^sub>t (Fun f X) \<A> Sec;
                  Fun f X \<in> GSMP\<^sub>s\<^sub>t (trms\<^sub>l\<^sub>s\<^sub>t \<A>)
                  \<rbrakk> \<Longrightarrow> P M (Fun f X)"
          "\<And>t K T t\<^sub>i. \<lbrakk>\<langle>M; \<A>; Sec\<rangle> \<turnstile>\<^sub>h\<^sub>o\<^sub>m t; P M t; Ana t = (K, T); \<And>k. k \<in> K \<Longrightarrow> \<langle>M; \<A>; Sec\<rangle> \<turnstile>\<^sub>h\<^sub>o\<^sub>m k;
                       \<And>k. k \<in> K \<Longrightarrow> P M k; t\<^sub>i \<in> T\<rbrakk> \<Longrightarrow> P M t\<^sub>i"
  shows "P M t"
using assms by (induct rule: intruder_deduct_hom.induct) blast+

lemma intruder_deduct_num_induct[consumes 1, case_names AxiomN ComposeN DecomposeN]:
  assumes "\<langle>M; n\<rangle> \<turnstile>\<^sub>n t" "\<And>t. t \<in> M \<Longrightarrow> P M 0 t"
          "\<And>T f steps.
              \<lbrakk>length T = arity f; public f;
               \<And>t. t \<in> set T \<Longrightarrow> \<langle>M; steps t\<rangle> \<turnstile>\<^sub>n t;
               \<And>t. t \<in> set T \<Longrightarrow> P M (steps t) t\<rbrakk>
              \<Longrightarrow> P M (Suc (Max (insert 0 (steps ` set T)))) (Fun f T)"
          "\<And>t K T t\<^sub>i steps n.
              \<lbrakk>\<langle>M; n\<rangle> \<turnstile>\<^sub>n t; P M n t; Ana t = (K, T);
               \<And>k. k \<in> K \<Longrightarrow> \<langle>M; steps k\<rangle> \<turnstile>\<^sub>n k;
               t\<^sub>i \<in> T; \<And>k. k \<in> K \<Longrightarrow> P M (steps k) k\<rbrakk>
              \<Longrightarrow> P M (Suc (Max (insert n (steps ` K)))) t\<^sub>i"
  shows "P M n t"
using assms by (induct rule: intruder_deduct_num.induct) blast+

lemma ideduct_hom_mono:
  "\<lbrakk>\<langle>M; \<A>; Sec\<rangle> \<turnstile>\<^sub>h\<^sub>o\<^sub>m t; M \<subseteq> M'\<rbrakk> \<Longrightarrow> \<langle>M'; \<A>; Sec\<rangle> \<turnstile>\<^sub>h\<^sub>o\<^sub>m t"
proof (induction rule: intruder_deduct_hom_induct)
  case (DecomposeH t K T t\<^sub>i)
  have "\<forall>k. k \<in> K \<longrightarrow> \<langle>M'; \<A>; Sec\<rangle> \<turnstile>\<^sub>h\<^sub>o\<^sub>m k" using DecomposeH.IH \<open>M \<subseteq> M'\<close> by simp
  moreover have "\<langle>M'; \<A>; Sec\<rangle> \<turnstile>\<^sub>h\<^sub>o\<^sub>m t" using DecomposeH.IH \<open>M \<subseteq> M'\<close> by simp
  ultimately show ?case
    using DecomposeH
          intruder_deduct_hom.DecomposeH[OF _ DecomposeH.hyps(2) _ DecomposeH.hyps(4)]
    by blast
qed auto


subsection {* Lemmata: Labeled constraint projections *}
lemma nil_lst_left_right_eq[simp]: "unlabel [] = []" "left_unl [] = []" "right_unl [] = []"
unfolding unlabel_def left_def right_def by auto

lemma singleton_lst_left_right[simp]:
  "left_unl [(l1, a)] = [a]" "left_unl [(l2, a)] = []" "left_unl [(\<star>, a)] = [a]"
  "right_unl [(l1, a)] = []" "right_unl [(l2, a)] = [a]" "right_unl [(\<star>, a)] = [a]"
  "unlabel [(l1, a)] = [a]" "unlabel [(l2, a)] = [a]" "unlabel [(\<star>, a)] = [a]"
unfolding left_def right_def unlabel_def by simp_all

lemma unlabel_nil_only_if_nil[simp]: "unlabel A = [] \<Longrightarrow> A = []"
unfolding unlabel_def by auto

lemma unlabel_step_append: "unlabel (a:A) = unlabel [a]:unlabel A"
unfolding unlabel_def by auto

lemma unlabel_cons[simp]: "unlabel ((l,a)#A) = a#unlabel A"
unfolding unlabel_def by simp

lemma unlabel_append: "unlabel (A:B) = unlabel A:unlabel B"
unfolding unlabel_def by auto

lemma lst_left_right_append:
  "left_unl (A:B) = left_unl A:left_unl B"
  "right_unl (A:B) = right_unl A:right_unl B"
unfolding left_def right_def unlabel_def by auto

lemma lst_left_unl_cons[simp]:
  "left_unl ((l1, a):A) = a:left_unl A"
  "left_unl ((l2, a):A) = left_unl A"
  "left_unl ((\<star>, a):A) = a:left_unl A"
unfolding left_def unlabel_def by simp_all

lemma lst_right_unl_cons[simp]:
  "right_unl ((l1, a):A) = right_unl A"
  "right_unl ((l2, a):A) = a:right_unl A"
  "right_unl ((\<star>, a):A) = a:right_unl A"
unfolding right_def unlabel_def by simp_all

lemma lst_unlabel_cons[simp]:
  "unlabel ((l1, a):A) = a:unlabel A"
  "unlabel ((l2, a):A) = a:unlabel A"
  "unlabel ((\<star>, a):A) = a:unlabel A"
unfolding unlabel_def by auto

lemma trms_unlabel_left[simp]:
  "trm\<^sub>s\<^sub>t\<^sub>p (snd (l1, x)) \<in> trms_left\<^sub>l\<^sub>s\<^sub>t [(l1, x)]"
  "trm_r\<^sub>s\<^sub>t\<^sub>p (snd (l1, x)) \<in> trms_left\<^sub>l\<^sub>s\<^sub>t [(l1, x)]"
by auto

lemma trms_unlabel_right[simp]:
  "trm\<^sub>s\<^sub>t\<^sub>p (snd (l2, x)) \<in> trms_right\<^sub>l\<^sub>s\<^sub>t [(l2, x)]"
  "trm_r\<^sub>s\<^sub>t\<^sub>p (snd (l2, x)) \<in> trms_right\<^sub>l\<^sub>s\<^sub>t [(l2, x)]"
by auto

lemma trms_unlabel_star[simp]:
  "trm\<^sub>s\<^sub>t\<^sub>p (snd (\<star>, x)) \<in> trms_left\<^sub>l\<^sub>s\<^sub>t [(\<star>, x)]"
  "trm_r\<^sub>s\<^sub>t\<^sub>p (snd (\<star>, x)) \<in> trms_left\<^sub>l\<^sub>s\<^sub>t [(\<star>, x)]"
  "trm\<^sub>s\<^sub>t\<^sub>p (snd (\<star>, x)) \<in> trms_right\<^sub>l\<^sub>s\<^sub>t [(\<star>, x)]"
  "trm_r\<^sub>s\<^sub>t\<^sub>p (snd (\<star>, x)) \<in> trms_right\<^sub>l\<^sub>s\<^sub>t [(\<star>, x)]"
by auto

lemma trms\<^sub>l\<^sub>s\<^sub>t_union[simp]: "trms\<^sub>l\<^sub>s\<^sub>t A = trms_left\<^sub>l\<^sub>s\<^sub>t A \<union> trms_right\<^sub>l\<^sub>s\<^sub>t A"
proof (induction A)
  case (Cons a A)
  obtain l s where "a = (l,s)" by moura
  thus ?case using Cons by (cases l) auto
qed simp

lemma trms\<^sub>l\<^sub>s\<^sub>t_append[simp]: "trms\<^sub>l\<^sub>s\<^sub>t (A@B) = trms\<^sub>l\<^sub>s\<^sub>t A \<union> trms\<^sub>l\<^sub>s\<^sub>t B"
proof (induction A)
  case (Cons a A)
  obtain l s where "a = (l,s)" by moura
  thus ?case using Cons by auto
qed simp

lemma trms_left\<^sub>l\<^sub>s\<^sub>t_append[simp]: "trms_left\<^sub>l\<^sub>s\<^sub>t (A@B) = trms_left\<^sub>l\<^sub>s\<^sub>t A \<union> trms_left\<^sub>l\<^sub>s\<^sub>t B"
proof (induction A)
  case (Cons a A)
  obtain l s where "a = (l,s)" by moura
  thus ?case using Cons by (cases l) auto
qed simp

lemma trms_right\<^sub>l\<^sub>s\<^sub>t_append[simp]: "trms_right\<^sub>l\<^sub>s\<^sub>t (A@B) = trms_right\<^sub>l\<^sub>s\<^sub>t A \<union> trms_right\<^sub>l\<^sub>s\<^sub>t B"
proof (induction A)
  case (Cons a A)
  obtain l s where "a = (l,s)" by moura
  thus ?case using Cons by (cases l) auto
qed simp

lemma trms_left\<^sub>l\<^sub>s\<^sub>t_subset[simp]:
  "trms_left\<^sub>l\<^sub>s\<^sub>t A \<subseteq> trms_left\<^sub>l\<^sub>s\<^sub>t (A@B)"
  "trms_left\<^sub>l\<^sub>s\<^sub>t B \<subseteq> trms_left\<^sub>l\<^sub>s\<^sub>t (A@B)"
proof (induction A)
  case (Cons a A)
  obtain l s where *: "a = (l,s)" by moura
  { case 1 thus ?case using Cons * by (cases l) auto }
  { case 2 thus ?case using Cons * by (cases l) auto }
qed simp_all

lemma trms_right\<^sub>l\<^sub>s\<^sub>t_subset[simp]:
  "trms_right\<^sub>l\<^sub>s\<^sub>t A \<subseteq> trms_right\<^sub>l\<^sub>s\<^sub>t (A@B)"
  "trms_right\<^sub>l\<^sub>s\<^sub>t B \<subseteq> trms_right\<^sub>l\<^sub>s\<^sub>t (A@B)"
proof (induction A)
  case (Cons a A)
  obtain l s where *: "a = (l,s)" by moura
  { case 1 thus ?case using Cons * by (cases l) auto }
  { case 2 thus ?case using Cons * by (cases l) auto }
qed simp_all

lemma trms\<^sub>l\<^sub>s\<^sub>t_subset[simp]:
  "trms\<^sub>l\<^sub>s\<^sub>t A \<subseteq> trms\<^sub>l\<^sub>s\<^sub>t (A@B)"
  "trms\<^sub>l\<^sub>s\<^sub>t B \<subseteq> trms\<^sub>l\<^sub>s\<^sub>t (A@B)"
proof (induction A)
  case (Cons a A)
  obtain l s where *: "a = (l,s)" by moura
  { case 1 thus ?case using Cons * by auto }
  { case 2 thus ?case using Cons * by auto }
qed simp_all

lemma vars\<^sub>l\<^sub>s\<^sub>t_union: "vars\<^sub>l\<^sub>s\<^sub>t A = vars_left\<^sub>l\<^sub>s\<^sub>t A \<union> vars_right\<^sub>l\<^sub>s\<^sub>t A"
proof (induction A)
  case (Cons a A)
  obtain l s where "a = (l,s)" by moura
  thus ?case using Cons by (cases l) auto
qed simp

lemma wfvarsoccs\<^sub>l\<^sub>s\<^sub>t_union:
  "wfvarsoccs\<^sub>s\<^sub>t (unlabel A) = wfvarsoccs\<^sub>s\<^sub>t (left_unl A) \<union> wfvarsoccs\<^sub>s\<^sub>t (right_unl A)"
proof (induction A)
  case (Cons a A)
  obtain l s where "a = (l,s)" by moura
  thus ?case using Cons by (cases l) auto
qed simp

lemma wf_both_if_wf_left_right:
  assumes "wf\<^sub>s\<^sub>t V (left_unl A)" "wf\<^sub>s\<^sub>t V (right_unl A)"
  shows "wf\<^sub>s\<^sub>t V (unlabel A)"
using assms
proof (induction A arbitrary: V rule: List.rev_induct)
  case (snoc a A)
  hence IH: "wf\<^sub>s\<^sub>t V (unlabel A)" using lst_left_right_append[of A] by auto
  obtain b where b: "a = (l1, b) \<or> a = (l2, b) \<or> a = (\<star>, b)"
    by (cases a, metis strand_label.exhaust)
  hence *: "wf\<^sub>s\<^sub>t V (left_unl A:b) \<or> wf\<^sub>s\<^sub>t V (right_unl A:b)"
    using snoc.prems lst_left_right_append[of A] by auto
  thus ?case using IH b snoc.prems lst_left_right_append[of A "[a]"] unlabel_append[of A "[a]"]
  proof (cases b)
    case (Receive t)
    hence "FV t \<subseteq> wfvarsoccs\<^sub>s\<^sub>t (unlabel A) \<union> V"
      using * b wf_append_exec[of V _ "[b]"] wfvarsoccs\<^sub>l\<^sub>s\<^sub>t_union[of A]
      by (cases a) auto
    hence "FV t \<subseteq> vars2\<^sub>s\<^sub>t (unlabel A) \<union> V"
      using vars_snd_rcv_strand_subset2(4)[of "unlabel A"]
      by auto
    hence "wf\<^sub>s\<^sub>t V (unlabel A:rcv\<langle>t\<rangle>)" by (rule wf_rcv_append'''[OF IH])
    thus ?thesis using b Receive unlabel_append[of A "[a]"] by auto
  next
    case (Equality s t)
    hence "FV t \<subseteq> wfvarsoccs\<^sub>s\<^sub>t (unlabel A) \<union> V"
      using * b wf_append_exec[of V _ "[b]"] wfvarsoccs\<^sub>l\<^sub>s\<^sub>t_union[of A]
      by (cases a) auto
    hence "FV t \<subseteq> vars2\<^sub>l\<^sub>s\<^sub>t A \<union> V"
      using vars_snd_rcv_strand_subset2(4)[of "unlabel A"]
      by auto
    hence "wf\<^sub>s\<^sub>t V (unlabel A:s \<doteq> t)" by (rule wf_eq_append'''[OF IH])
    thus ?thesis using b Equality unlabel_append[of A "[a]"] by auto
  qed auto
qed simp

lemma unlabel_cons_inv:
  "unlabel A = b#B \<Longrightarrow> \<exists>A'. A = (l1, b)#A' \<or> A = (l2, b)#A' \<or> A = (\<star>, b)#A'"
unfolding unlabel_def by (auto intro: strand_label.exhaust)

lemma unlabel_snoc_inv:
  "unlabel A = B@[b] \<Longrightarrow> \<exists>A'. A = A'@[(l1, b)] \<or> A = A'@[(l2, b)] \<or> A = A'@[(\<star>, b)]"
proof -
  assume *: "unlabel A = B@[b]"
  then obtain A' l where "A = A'@[(l,b)]"
    unfolding unlabel_def by (induct A rule: List.rev_induct) auto
  thus "\<exists>A'. A = A'@[(l1, b)] \<or> A = A'@[(l2, b)] \<or> A = A'@[(\<star>, b)]" by (cases l) auto
qed

lemma left_right_idem: "left (left A) = left A" "right (right A) = right A"
unfolding left_def right_def by auto

lemma left_right_ik_union_is_unlabel_ik: "ik\<^sub>s\<^sub>t (left_unl A) \<union> ik\<^sub>s\<^sub>t (right_unl A) = ik\<^sub>s\<^sub>t (unlabel A)"
proof (induction "unlabel A" arbitrary: A)
  case (Nil A)
  hence "A = []" by simp
  thus ?case by simp
next
  case (Cons a A' B)
  then obtain B' where B': "B = (l1, a)#B' \<or> B = (l2, a)#B' \<or> B = (\<star>, a)#B'"
    by (metis unlabel_cons_inv)
  hence "A' = unlabel B'" using Cons.hyps(2) by (cases B) auto
  hence "ik\<^sub>s\<^sub>t (left_unl B') \<union> ik\<^sub>s\<^sub>t (right_unl B') = ik\<^sub>s\<^sub>t (unlabel B')" by (metis Cons.hyps(1))
  thus ?case using B' by (cases a) auto
qed


subsection {* Lemmata: GSMP *}

lemma GSMP\<^sub>s\<^sub>t_subterm:
  assumes "t \<in> GSMP\<^sub>s\<^sub>t M" "t' \<sqsubseteq> t"
  shows "t' \<in> GSMP\<^sub>s\<^sub>t M"
using SMP\<^sub>s\<^sub>t.Subterm[of t M t'] ground_subterm[of t t'] assms by auto

lemma GSMP\<^sub>s\<^sub>t_Ana_key:
  assumes "t \<in> GSMP\<^sub>s\<^sub>t M" "Ana t = (K,T)" "k \<in> K"
  shows "k \<in> GSMP\<^sub>s\<^sub>t M"
using SMP\<^sub>s\<^sub>t.Ana[of t M K T k] Ana_keys_FV[of t K T] assms by auto

lemma GSMP_append[simp]: "GSMP\<^sub>s\<^sub>t (trms\<^sub>l\<^sub>s\<^sub>t (A@B)) = GSMP\<^sub>s\<^sub>t (trms\<^sub>l\<^sub>s\<^sub>t A) \<union> GSMP\<^sub>s\<^sub>t (trms\<^sub>l\<^sub>s\<^sub>t B)"
using SMP_append'[of "trms\<^sub>l\<^sub>s\<^sub>t A" "trms\<^sub>l\<^sub>s\<^sub>t B"] trms\<^sub>l\<^sub>s\<^sub>t_append[of A B] by auto

lemma GSMP_append': "GSMP\<^sub>s\<^sub>t (A \<union> B) = GSMP\<^sub>s\<^sub>t A \<union> GSMP\<^sub>s\<^sub>t B"
using SMP_append'[of A B] by auto

lemma GSMP_union: "GSMP\<^sub>s\<^sub>t (trms\<^sub>l\<^sub>s\<^sub>t A) = GSMP\<^sub>s\<^sub>t (trms_left\<^sub>l\<^sub>s\<^sub>t A) \<union> GSMP\<^sub>s\<^sub>t (trms_right\<^sub>l\<^sub>s\<^sub>t A)"
using trms\<^sub>l\<^sub>s\<^sub>t_union[of A] SMP_append'[of "trms_left\<^sub>l\<^sub>s\<^sub>t A" "trms_right\<^sub>l\<^sub>s\<^sub>t A"] by auto

lemma GSMP_disjoint_term:
  assumes "GSMP_disjoint (trms_left\<^sub>l\<^sub>s\<^sub>t \<A>) (trms_right\<^sub>l\<^sub>s\<^sub>t \<A>) Sec"
  shows "t \<notin> GSMP\<^sub>s\<^sub>t (trms_left\<^sub>l\<^sub>s\<^sub>t \<A>) \<or> t \<notin> GSMP\<^sub>s\<^sub>t (trms_right\<^sub>l\<^sub>s\<^sub>t \<A>) \<or> t \<in> Sec \<or> {} \<turnstile>\<^sub>c t"
using assms unfolding GSMP_disjoint_def by auto

lemma GSMP_wt_subst_subset:
  assumes "t \<in> GSMP\<^sub>s\<^sub>t (M \<cdot>set \<I>)" "wt\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<I>" "\<forall>x. wf\<^sub>t\<^sub>r\<^sub>m (\<I> x)"
  shows "t \<in> GSMP\<^sub>s\<^sub>t M"
using SMP_wt_subst_subset[OF _ assms(2,3), of t M] assms(1) by simp

lemma GSMP_wt_substI:
  assumes "t \<in> M" "wt\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t I" "\<forall>x. wf\<^sub>t\<^sub>r\<^sub>m (I x)" "interpretation\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t I"
  shows "t \<cdot> I \<in> GSMP\<^sub>s\<^sub>t M"
proof -
  have "t \<in> SMP\<^sub>s\<^sub>t M" using assms(1) by auto
  hence *: "t \<cdot> I \<in> SMP\<^sub>s\<^sub>t M" using SMP\<^sub>s\<^sub>t.Substitution assms(2,3) by simp
  moreover have "FV (t \<cdot> I) = {}" using assms(1) interpretation_grounds_all'[OF assms(4)] by auto
  ultimately show ?thesis by simp
qed

lemma GSMP_disjoint_subset:
  assumes "GSMP_disjoint L R S" "L' \<subseteq> L" "R' \<subseteq> R"
  shows "GSMP_disjoint L' R' S"
using assms(1) SMP_mono[OF assms(2)] SMP_mono[OF assms(3)]
by auto

lemma GSMP_disjoint_left_specific_not_right_specific:
  assumes "GSMP_disjoint (trms_left\<^sub>l\<^sub>s\<^sub>t \<A>) (trms_right\<^sub>l\<^sub>s\<^sub>t \<A>) Sec"
  and "left_specific m \<A> Sec"
  shows "\<not>right_specific m \<A> Sec"
using assms by auto

lemma GSMP_disjoint_right_specific_not_left_specific:
  assumes "GSMP_disjoint (trms_left\<^sub>l\<^sub>s\<^sub>t \<A>) (trms_right\<^sub>l\<^sub>s\<^sub>t \<A>) Sec"
  and "right_specific m \<A> Sec"
  shows "\<not>left_specific m \<A> Sec"
using assms by auto

lemma GSMP_disjoint_intersection_not_specific:
  assumes "GSMP_disjoint (trms_left\<^sub>l\<^sub>s\<^sub>t \<A>) (trms_right\<^sub>l\<^sub>s\<^sub>t \<A>) Sec"
  and "t \<in> Sec \<or> {} \<turnstile>\<^sub>c t"
  shows "\<not>left_specific t \<A> Sec" "\<not>right_specific t \<A> Sec"
using assms by auto


subsection {* Lemmata: Homogeneous and heterogeneous terms *}

lemma heterogeneous_supterm:
  assumes "heterogeneous\<^sub>l\<^sub>s\<^sub>t t \<A> Sec" "t \<sqsubseteq> t'"
  shows "heterogeneous\<^sub>l\<^sub>s\<^sub>t t' \<A> Sec"
using assms subtermeq_trans by auto

lemma homogeneous_subterm:
  assumes "homogeneous\<^sub>l\<^sub>s\<^sub>t t \<A> Sec" "t' \<sqsubseteq> t"
  shows "homogeneous\<^sub>l\<^sub>s\<^sub>t t' \<A> Sec"
by (metis assms heterogeneous_supterm)

lemma left_specific_subterm:
  assumes "t \<sqsubseteq> t'" "left_specific t' \<A> Sec"
  shows "left_specific t \<A> Sec \<or> t \<in> Sec \<or> {} \<turnstile>\<^sub>c t"
using GSMP\<^sub>s\<^sub>t_subterm[OF _ assms(1)] assms(2) by auto

lemma right_specific_subterm:
  assumes "t \<sqsubseteq> t'" "right_specific t' \<A> Sec"
  shows "right_specific t \<A> Sec \<or> t \<in> Sec \<or> {} \<turnstile>\<^sub>c t"
using GSMP\<^sub>s\<^sub>t_subterm[OF _ assms(1)] assms(2) by auto

lemma heterogeneous_term_is_Fun:
  assumes "heterogeneous\<^sub>l\<^sub>s\<^sub>t t A S" shows "\<exists>f T. t = Fun f T"
using assms by (cases t) auto

lemma left_specific_is_homogeneous:
  assumes \<A>: "GSMP_disjoint (trms_left\<^sub>l\<^sub>s\<^sub>t \<A>) (trms_right\<^sub>l\<^sub>s\<^sub>t \<A>) Sec"
  and t: "left_specific m \<A> Sec"
  shows "homogeneous\<^sub>l\<^sub>s\<^sub>t m \<A> Sec"
proof
  assume "heterogeneous\<^sub>l\<^sub>s\<^sub>t m \<A> Sec"
  then obtain s where s: "s \<in> subterms m" "right_specific s \<A> Sec" by moura
  hence "s \<in> GSMP\<^sub>s\<^sub>t (trms_left\<^sub>l\<^sub>s\<^sub>t \<A>)" "s \<in> GSMP\<^sub>s\<^sub>t (trms_right\<^sub>l\<^sub>s\<^sub>t \<A>)" using t by auto
  hence "s \<in> Sec \<or> {} \<turnstile>\<^sub>c s" using \<A> by auto
  thus False using s(2) by auto
qed

lemma right_specific_is_homogeneous:
  assumes \<A>: "GSMP_disjoint (trms_left\<^sub>l\<^sub>s\<^sub>t \<A>) (trms_right\<^sub>l\<^sub>s\<^sub>t \<A>) Sec"
  and t: "right_specific m \<A> Sec"
  shows "homogeneous\<^sub>l\<^sub>s\<^sub>t m \<A> Sec"
proof
  assume "heterogeneous\<^sub>l\<^sub>s\<^sub>t m \<A> Sec"
  then obtain s where s: "s \<in> subterms m" "left_specific s \<A> Sec" by moura
  hence "s \<in> GSMP\<^sub>s\<^sub>t (trms_left\<^sub>l\<^sub>s\<^sub>t \<A>)" "s \<in> GSMP\<^sub>s\<^sub>t (trms_right\<^sub>l\<^sub>s\<^sub>t \<A>)" using t by auto
  hence "s \<in> Sec \<or> {} \<turnstile>\<^sub>c s" using \<A> by auto
  thus False using s(2) by auto
qed

lemma deduct_synth_homogeneous:
  assumes "{} \<turnstile>\<^sub>c t"
  shows "homogeneous\<^sub>l\<^sub>s\<^sub>t t \<A> Sec"
proof -
  have "\<forall>s \<in> subterms t. {} \<turnstile>\<^sub>c s" using deduct_synth_subterm[OF assms] by auto
  thus ?thesis unfolding heterogeneous\<^sub>l\<^sub>s\<^sub>t_def left_specific_def right_specific_def by auto
qed

lemma GSMP_left_is_homogeneous:
  assumes "GSMP_disjoint (trms_left\<^sub>l\<^sub>s\<^sub>t A) (trms_right\<^sub>l\<^sub>s\<^sub>t A) Sec"
  and "t \<in> GSMP\<^sub>s\<^sub>t (trms_left\<^sub>l\<^sub>s\<^sub>t A)" "t \<notin> Sec"
  shows "homogeneous\<^sub>l\<^sub>s\<^sub>t t A Sec"
using assms left_specific_is_homogeneous deduct_synth_homogeneous
by (cases "{} \<turnstile>\<^sub>c t") auto

lemma GSMP_right_is_homogeneous:
  assumes "GSMP_disjoint (trms_left\<^sub>l\<^sub>s\<^sub>t A) (trms_right\<^sub>l\<^sub>s\<^sub>t A) Sec"
  and "t \<in> GSMP\<^sub>s\<^sub>t (trms_right\<^sub>l\<^sub>s\<^sub>t A)" "t \<notin> Sec"
  shows "homogeneous\<^sub>l\<^sub>s\<^sub>t t A Sec"
using assms right_specific_is_homogeneous deduct_synth_homogeneous
by (cases "{} \<turnstile>\<^sub>c t") auto

lemma homogeneous_is_not_left_right_specific:
  assumes "homogeneous\<^sub>l\<^sub>s\<^sub>t m \<A> Sec"
  shows "\<not>left_specific m \<A> Sec \<or> \<not>right_specific m \<A> Sec"
using assms by auto

lemma secrets_are_homogeneous:
  assumes "\<forall>s \<in> Sec. \<forall>s' \<in> subterms s. {} \<turnstile>\<^sub>c s' \<or> s' \<in> Sec" "s \<in> Sec"
  shows "homogeneous\<^sub>l\<^sub>s\<^sub>t s \<A> Sec"
using assms unfolding heterogeneous\<^sub>l\<^sub>s\<^sub>t_def left_specific_def right_specific_def
using assms by auto

lemma GSMP_is_homogeneous:
  assumes \<A>: "GSMP_disjoint (trms_left\<^sub>l\<^sub>s\<^sub>t \<A>) (trms_right\<^sub>l\<^sub>s\<^sub>t \<A>) Sec"
  and Sec: "\<forall>s \<in> Sec. \<forall>s' \<in> subterms s. {} \<turnstile>\<^sub>c s' \<or> s' \<in> Sec"
  and t: "t \<in> GSMP\<^sub>s\<^sub>t (trms\<^sub>l\<^sub>s\<^sub>t \<A>)"
  shows "homogeneous\<^sub>l\<^sub>s\<^sub>t t \<A> Sec"
proof -
  have "t \<in> GSMP\<^sub>s\<^sub>t (trms_left\<^sub>l\<^sub>s\<^sub>t \<A>) \<or> t \<in> GSMP\<^sub>s\<^sub>t (trms_right\<^sub>l\<^sub>s\<^sub>t \<A>)"
    using t GSMP_append'[of "trms_left\<^sub>l\<^sub>s\<^sub>t \<A>" "trms_right\<^sub>l\<^sub>s\<^sub>t \<A>"] trms\<^sub>l\<^sub>s\<^sub>t_union[of \<A>]
    by auto
  thus ?thesis
    using GSMP_left_is_homogeneous[OF \<A>, of t]
          GSMP_right_is_homogeneous[OF \<A>, of t]
          secrets_are_homogeneous[OF Sec, of t]
    by auto
qed

lemma Ana_keys_homogeneous:
  assumes \<A>: "GSMP_disjoint (trms_left\<^sub>l\<^sub>s\<^sub>t \<A>) (trms_right\<^sub>l\<^sub>s\<^sub>t \<A>) Sec"
  and Sec: "\<forall>s \<in> Sec. \<forall>s' \<in> subterms s. {} \<turnstile>\<^sub>c s' \<or> s' \<in> Sec"
  and t: "t \<in> GSMP\<^sub>s\<^sub>t (trms\<^sub>l\<^sub>s\<^sub>t \<A>)"
  and k: "Ana t = (K,T)" "k \<in> K"
  shows "homogeneous\<^sub>l\<^sub>s\<^sub>t k \<A> Sec"
proof -
  from t k have "k \<in> GSMP\<^sub>s\<^sub>t (trms\<^sub>l\<^sub>s\<^sub>t \<A>)"
    using SMP\<^sub>s\<^sub>t.Ana[OF _ k(1,2)] Ana_keys_FV[OF k(1)] k(2)
    by auto
  thus ?thesis using GSMP_is_homogeneous[OF \<A> Sec] by auto
qed


subsection {* Lemmata: Intruder knowledge and declassification *}

lemma ik_left_subset[simp]: "ik\<^sub>s\<^sub>t (left_unl A) \<subseteq> trms_left\<^sub>l\<^sub>s\<^sub>t A"
by auto

lemma ik_right_subset[simp]: "ik\<^sub>s\<^sub>t (right_unl A) \<subseteq> trms_right\<^sub>l\<^sub>s\<^sub>t A"
by auto

lemma ik_left_union[simp]:
  "ik\<^sub>s\<^sub>t (left_unl (A@B)) = ik\<^sub>s\<^sub>t (left_unl A) \<union> ik\<^sub>s\<^sub>t (left_unl B)"
using lst_left_right_append[of A B] by auto

lemma ik_right_union[simp]:
  "ik\<^sub>s\<^sub>t (right_unl (A@B)) = ik\<^sub>s\<^sub>t (right_unl A) \<union> ik\<^sub>s\<^sub>t (right_unl B)"
using lst_left_right_append[of A B] by auto

lemma ik_left_subst_GSMP_subset:
  assumes I: "wt\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t I" "\<forall>x. wf\<^sub>t\<^sub>r\<^sub>m (I x)" "interpretation\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t I"
  shows "ik\<^sub>s\<^sub>t (left_unl A) \<cdot>set I \<subseteq> GSMP\<^sub>s\<^sub>t (trms_left\<^sub>l\<^sub>s\<^sub>t A)"
proof
  fix t assume "t \<in> ik\<^sub>s\<^sub>t (left_unl A) \<cdot>set I"
  hence *: "t \<in> trms_left\<^sub>l\<^sub>s\<^sub>t A \<cdot>set I" by auto
  then obtain s where "s \<in> trms_left\<^sub>l\<^sub>s\<^sub>t A" "t = s \<cdot> I" by auto
  hence "t \<in> SMP\<^sub>s\<^sub>t (trms_left\<^sub>l\<^sub>s\<^sub>t A)" using SMP_I I(1,2) by simp
  moreover have "FV t = {}" using * interpretation_grounds_all'[OF I(3)] by auto
  ultimately show "t \<in> GSMP\<^sub>s\<^sub>t (trms_left\<^sub>l\<^sub>s\<^sub>t A)" by simp
qed

lemma ik_right_subst_GSMP_subset:
  assumes I: "wt\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t I" "\<forall>x. wf\<^sub>t\<^sub>r\<^sub>m (I x)" "interpretation\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t I"
  shows "ik\<^sub>s\<^sub>t (right_unl A) \<cdot>set I \<subseteq> GSMP\<^sub>s\<^sub>t (trms_right\<^sub>l\<^sub>s\<^sub>t A)"
proof
  fix t assume "t \<in> ik\<^sub>s\<^sub>t (right_unl A) \<cdot>set I"
  hence *: "t \<in> trms_right\<^sub>l\<^sub>s\<^sub>t A \<cdot>set I" by auto
  then obtain s where "s \<in> trms_right\<^sub>l\<^sub>s\<^sub>t A" "t = s \<cdot> I" by auto
  hence "t \<in> SMP\<^sub>s\<^sub>t (trms_right\<^sub>l\<^sub>s\<^sub>t A)" using SMP_I I(1,2) by simp
  moreover have "FV t = {}" using * interpretation_grounds_all'[OF I(3)] by auto
  ultimately show "t \<in> GSMP\<^sub>s\<^sub>t (trms_right\<^sub>l\<^sub>s\<^sub>t A)" by simp
qed

lemma declassified_left_ik_subset: "declassified\<^sub>l\<^sub>s\<^sub>t A I \<subseteq> ik\<^sub>s\<^sub>t (left_unl A) \<cdot>set I"
proof (induction A)
  case (Cons a A) thus ?case using ik_left_union[of "[a]" A] by auto
qed simp

lemma declassified_right_ik_subset: "declassified\<^sub>l\<^sub>s\<^sub>t A I \<subseteq> ik\<^sub>s\<^sub>t (right_unl A) \<cdot>set I"
proof (induction A)
  case (Cons a A) thus ?case using ik_right_union[of "[a]" A] by auto
qed simp

lemma declassified_left_GSMP_subset:
  assumes I: "wt\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t I" "\<forall>x. wf\<^sub>t\<^sub>r\<^sub>m (I x)" "interpretation\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t I"
  shows "declassified\<^sub>l\<^sub>s\<^sub>t A I \<subseteq> GSMP\<^sub>s\<^sub>t (trms_left\<^sub>l\<^sub>s\<^sub>t A)"
by (rule subset_trans[OF declassified_left_ik_subset ik_left_subst_GSMP_subset[OF I]])

lemma declassified_right_GSMP_subset:
  assumes I: "wt\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t I" "\<forall>x. wf\<^sub>t\<^sub>r\<^sub>m (I x)" "interpretation\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t I"
  shows "declassified\<^sub>l\<^sub>s\<^sub>t A I \<subseteq> GSMP\<^sub>s\<^sub>t (trms_right\<^sub>l\<^sub>s\<^sub>t A)"
by (rule subset_trans[OF declassified_right_ik_subset ik_right_subst_GSMP_subset[OF I]])

lemma declassified_secrets_subset:
  assumes A: "GSMP_disjoint (trms_left\<^sub>l\<^sub>s\<^sub>t A) (trms_right\<^sub>l\<^sub>s\<^sub>t A) Sec"
  and I: "wt\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t I" "\<forall>x. wf\<^sub>t\<^sub>r\<^sub>m (I x)" "interpretation\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t I"
  shows "declassified\<^sub>l\<^sub>s\<^sub>t A I \<subseteq> Sec \<union> {m. {} \<turnstile>\<^sub>c m}"
using declassified_left_GSMP_subset[OF I] declassified_right_GSMP_subset[OF I] A
unfolding GSMP_disjoint_def by blast


subsection {* Lemmata: Intruder deduction equivalences *}

lemma deduct_if_hom_deduct: "\<langle>M;A;S\<rangle> \<turnstile>\<^sub>h\<^sub>o\<^sub>m m \<Longrightarrow> M \<turnstile> m"
proof (induction m rule: intruder_deduct_hom.induct)
  case (DecomposeH M A S t K T t\<^sub>i) thus ?case using intruder_deduct.Decompose by blast
qed simp_all

lemma hom_deduct_if_hom_ik:
  assumes "\<langle>M;A;Sec\<rangle> \<turnstile>\<^sub>h\<^sub>o\<^sub>m m" "\<forall>m \<in> M. homogeneous\<^sub>l\<^sub>s\<^sub>t m A Sec \<and> m \<in> GSMP\<^sub>s\<^sub>t (trms\<^sub>l\<^sub>s\<^sub>t A)"
  shows "homogeneous\<^sub>l\<^sub>s\<^sub>t m A Sec \<and> m \<in> GSMP\<^sub>s\<^sub>t (trms\<^sub>l\<^sub>s\<^sub>t A)"
using assms
proof (induction m rule: intruder_deduct_hom_induct)
  case (DecomposeH t K T t\<^sub>i)
  obtain f S where "t = Fun f S" using Ana_var \<open>t\<^sub>i \<in> T\<close> \<open>Ana t = (K, T)\<close> by (cases t) auto
  thus ?case
    using DecomposeH
          Ana_fun_subterm[of f S K T]
          homogeneous_subterm[OF _ subtermeqI'', of f S A Sec t\<^sub>i]
          GSMP\<^sub>s\<^sub>t_subterm[OF _ subtermeqI'', of f S "trms\<^sub>l\<^sub>s\<^sub>t A" t\<^sub>i]
    by blast
qed simp_all

lemma deduct_hom_if_synth:
  assumes hom: "homogeneous\<^sub>l\<^sub>s\<^sub>t m \<A> Sec" "m \<in> GSMP\<^sub>s\<^sub>t (trms\<^sub>l\<^sub>s\<^sub>t \<A>)"
  and m: "M \<turnstile>\<^sub>c m"
  shows "\<langle>M; \<A>; Sec\<rangle> \<turnstile>\<^sub>h\<^sub>o\<^sub>m m"
using m hom
proof (induction m rule: intruder_synth_induct)
  case (ComposeC T f)
  hence "\<And>x. x \<in> set T \<Longrightarrow> \<langle>M; \<A>; Sec\<rangle> \<turnstile>\<^sub>h\<^sub>o\<^sub>m x"
    using homogeneous_subterm[OF _ subtermeqI'', of f T \<A> Sec]
          GSMP\<^sub>s\<^sub>t_subterm[OF _ subtermeqI'', of f T "trms\<^sub>l\<^sub>s\<^sub>t \<A>"]
    by fastforce
  thus ?case by (metis intruder_deduct_hom.ComposeH[OF ComposeC.hyps(1,2) _ ComposeC.prems(1,2)])
qed simp

lemma declassified_append: "declassified\<^sub>l\<^sub>s\<^sub>t (A@B) I = declassified\<^sub>l\<^sub>s\<^sub>t A I \<union> declassified\<^sub>l\<^sub>s\<^sub>t B I"
using assms unfolding declassified\<^sub>l\<^sub>s\<^sub>t_def by auto

lemma deduct_zero_in_ik:
  assumes "\<langle>M; 0\<rangle> \<turnstile>\<^sub>n t" shows "t \<in> M"
proof -
  { fix k assume "\<langle>M; k\<rangle> \<turnstile>\<^sub>n t" hence "k > 0 \<or> t \<in> M" by (induct t) auto
  } thus ?thesis using assms by auto
qed

lemma deduct_if_deduct_num: "\<langle>M; k\<rangle> \<turnstile>\<^sub>n t \<Longrightarrow> M \<turnstile> t"
by (induct t rule: intruder_deduct_num.induct)
   (metis intruder_deduct.Axiom,
    metis intruder_deduct.Compose,
    metis intruder_deduct.Decompose)

lemma deduct_num_if_deduct: "M \<turnstile> t \<Longrightarrow> \<exists>k. \<langle>M; k\<rangle> \<turnstile>\<^sub>n t"
proof (induction t rule: intruder_deduct_induct)
  case (Compose T f)
  then obtain steps where *: "\<forall>t \<in> set T. \<langle>M; steps t\<rangle> \<turnstile>\<^sub>n t" by moura
  then obtain n where "\<forall>t \<in> set T. steps t \<le> n"
    using finite_nat_set_iff_bounded_le[of "steps ` set T"]
    by auto
  thus ?case using ComposeN[OF Compose.hyps(1,2), of M steps] * by force
next
  case (Decompose t K T t\<^sub>i)
  hence "\<And>u. u \<in> insert t K \<Longrightarrow> \<exists>k. \<langle>M; k\<rangle> \<turnstile>\<^sub>n u" by auto
  then obtain steps where *: "\<langle>M; steps t\<rangle> \<turnstile>\<^sub>n t" "\<forall>t \<in> K. \<langle>M; steps t\<rangle> \<turnstile>\<^sub>n t" by moura
  then obtain n where "steps t \<le> n" "\<forall>t \<in> K. steps t \<le> n"
    using finite_nat_set_iff_bounded_le[of "steps ` insert t K"]
          Ana_keys_finite[OF Decompose.hyps(2)]
    by auto
  thus ?case using DecomposeN[OF _ Decompose.hyps(2) _ Decompose.hyps(4), of M _ steps] * by force
qed (metis AxiomN)

lemma deduct_normalize:
  assumes M: "\<forall>m \<in> M. \<forall>f T. Fun f T \<sqsubseteq> m \<longrightarrow> P (Fun f T)"
  and t: "\<langle>M; k\<rangle> \<turnstile>\<^sub>n t" "Fun f T \<sqsubseteq> t" "\<not>P (Fun f T)"
  shows "(\<exists>l \<le> k. \<langle>M; l\<rangle> \<turnstile>\<^sub>n Fun f T) \<and> (\<forall>t \<in> set T. \<exists>l < k. \<langle>M; l\<rangle> \<turnstile>\<^sub>n t)"
using t
proof (induction t rule: intruder_deduct_num_induct)
  case (AxiomN t) thus ?case using M by auto
next
  case (ComposeN T' f' steps) thus ?case
  proof (cases "Fun f' T' = Fun f T")
    case True
    hence "\<langle>M; Suc (Max (insert 0 (steps ` set T')))\<rangle> \<turnstile>\<^sub>n Fun f T" "T = T'"
      using intruder_deduct_num.ComposeN[OF ComposeN.hyps] by auto
    moreover have "\<And>t. t \<in> set T \<Longrightarrow> \<langle>M; steps t\<rangle> \<turnstile>\<^sub>n t"
      using True ComposeN.hyps(3) by auto
    moreover have "\<And>t. t \<in> set T \<Longrightarrow> steps t < Suc (Max (insert 0 (steps ` set T)))"
      using Max_less_iff[of "insert 0 (steps ` set T)" "Suc (Max (insert 0 (steps ` set T)))"]
      by auto
    ultimately show ?thesis by auto
  next
    case False
    then obtain t' where t': "t' \<in> set T'" "Fun f T \<sqsubseteq> t'" using ComposeN by auto
    hence "(\<exists>l \<le> steps t'. \<langle>M; l\<rangle> \<turnstile>\<^sub>n Fun f T) \<and>
           (\<forall>t \<in> set T. \<exists>l < steps t'. \<langle>M; l\<rangle> \<turnstile>\<^sub>n t)"
      using ComposeN.IH[OF _ _ ComposeN.prems(2)] by auto
    moreover have "steps t' < Suc (Max (insert 0 (steps ` set T')))"
      using Max_less_iff[of "insert 0 (steps ` set T')" "Suc (Max (insert 0 (steps ` set T')))"]
      using t'(1) by auto
    ultimately show ?thesis using ComposeN.hyps(3)[OF t'(1)]
      by (meson Suc_le_eq le_Suc_eq le_trans)
  qed
next
  case (DecomposeN t K T' t\<^sub>i steps n)
  hence *: "Fun f T \<sqsubseteq> t"
    using subtermeq_trans[of "Fun f T" t\<^sub>i t] Ana_subterm[of t K T'] subtermeqI[of t\<^sub>i t]
    by blast
  have "(\<exists>l \<le> n. \<langle>M; l\<rangle> \<turnstile>\<^sub>n Fun f T) \<and> (\<forall>t' \<in> set T. \<exists>l < n. \<langle>M; l\<rangle> \<turnstile>\<^sub>n t')"
    using DecomposeN.IH(1)[OF * DecomposeN.prems(2)] by auto
  moreover have "n < Suc (Max (insert n (steps ` K)))"
      using Max_less_iff[of "insert n (steps ` K)" "Suc (Max (insert n (steps ` K)))"]
      using Ana_keys_finite[OF DecomposeN.hyps(2)] by auto
  ultimately show ?case using DecomposeN.hyps(4) by (meson Suc_le_eq le_Suc_eq le_trans)
qed

lemma hom_deduct_if_deduct:
  assumes \<A>: "par_comp \<A> Sec"
  and M: "\<forall>m\<in>M. homogeneous\<^sub>l\<^sub>s\<^sub>t m \<A> Sec \<and> m \<in> GSMP\<^sub>s\<^sub>t (trms\<^sub>l\<^sub>s\<^sub>t \<A>)"
  and m: "M \<turnstile> m" "homogeneous\<^sub>l\<^sub>s\<^sub>t m \<A> Sec" "m \<in> GSMP\<^sub>s\<^sub>t (trms\<^sub>l\<^sub>s\<^sub>t \<A>)"
  shows "\<langle>M; \<A>; Sec\<rangle> \<turnstile>\<^sub>h\<^sub>o\<^sub>m m"
proof -
  let ?P = "\<lambda>x. homogeneous\<^sub>l\<^sub>s\<^sub>t x \<A> Sec \<and> x \<in> GSMP\<^sub>s\<^sub>t (trms\<^sub>l\<^sub>s\<^sub>t \<A>)"
  { fix k assume "\<langle>M; k\<rangle> \<turnstile>\<^sub>n m"
    hence ?thesis using m(2,3)
    proof (induction k arbitrary: m rule: nat_less_induct)
      case (1 n m) thus ?case
      proof (cases n)
        case 0
        hence "m \<in> M" using deduct_zero_in_ik "1.prems"(1) by metis
        thus ?thesis by auto
      next
        case (Suc n')
        hence "\<langle>M; Suc n'\<rangle> \<turnstile>\<^sub>n m"
              "\<forall>m<Suc n'. \<forall>x. \<langle>M; m\<rangle> \<turnstile>\<^sub>n x \<longrightarrow> ?P x \<longrightarrow> \<langle>M;\<A>;Sec\<rangle> \<turnstile>\<^sub>h\<^sub>o\<^sub>m x"
          using "1.prems" "1.IH" by meson+
        thus ?thesis using "1.prems"(2,3)
        proof (induction m rule: intruder_deduct_num_induct)
          case (ComposeN T f steps)
          have "\<And>x. x \<in> set T \<Longrightarrow> ?P x"
                "\<And>x. x \<in> set T \<Longrightarrow> steps x < Suc (Max (insert 0 (steps ` set T)))"
            using homogeneous_subterm[OF _ subtermeqI'', of f T \<A> Sec]
                  GSMP\<^sub>s\<^sub>t_subterm[OF _ subtermeqI'', of f T "trms\<^sub>l\<^sub>s\<^sub>t \<A>"]
                  ComposeN.prems(2,3)
                  Max_less_iff[of "insert 0 (steps ` set T)" "Suc (Max (insert 0 (steps ` set T)))"]
            by auto
          hence "\<And>x. x \<in> set T \<Longrightarrow> \<langle>M; \<A>; Sec\<rangle> \<turnstile>\<^sub>h\<^sub>o\<^sub>m x"
            using ComposeN.prems(1) ComposeN.hyps(3) by blast
          thus ?case
            by (metis intruder_deduct_hom.ComposeH[OF ComposeN.hyps(1,2) _ ComposeN.prems(2,3)])
        next
          case (DecomposeN t K T t\<^sub>i steps l)
          have disj: "GSMP_disjoint (trms_left\<^sub>l\<^sub>s\<^sub>t \<A>) (trms_right\<^sub>l\<^sub>s\<^sub>t \<A>) Sec" using \<A> by simp
          show ?case
          proof (cases "?P t")
            case True
            hence "\<And>k. k \<in> K \<Longrightarrow> ?P k"
              using Ana_keys_homogeneous[OF disj _ _ DecomposeN.hyps(2)] \<A>
                    GSMP\<^sub>s\<^sub>t_Ana_key[OF _ DecomposeN.hyps(2), of "trms\<^sub>l\<^sub>s\<^sub>t \<A>"]
              by auto
            moreover have
                "\<And>k m x. k \<in> K \<Longrightarrow> m < steps k \<Longrightarrow> \<langle>M; m\<rangle> \<turnstile>\<^sub>n x \<Longrightarrow> ?P x \<Longrightarrow> \<langle>M;\<A>;Sec\<rangle> \<turnstile>\<^sub>h\<^sub>o\<^sub>m x"
            proof -
              fix k m x assume *: "k \<in> K" "m < steps k" "\<langle>M; m\<rangle> \<turnstile>\<^sub>n x" "?P x"
              have "steps k \<in> insert l (steps ` K)" using *(1) by simp
              hence "m < Suc (Max (insert l (steps ` K)))"
                using less_trans[OF *(2), of "Suc (Max (insert l (steps ` K)))"]
                      Max_less_iff[of "insert l (steps ` K)" "Suc (Max (insert l (steps ` K)))"]
                      Ana_keys_finite[OF DecomposeN.hyps(2)]
                by auto
              thus "\<langle>M;\<A>;Sec\<rangle> \<turnstile>\<^sub>h\<^sub>o\<^sub>m x" using DecomposeN.prems(1) *(3,4) by simp
            qed
            ultimately have "\<And>k. k \<in> K \<Longrightarrow> \<langle>M; \<A>; Sec\<rangle> \<turnstile>\<^sub>h\<^sub>o\<^sub>m k"
              using DecomposeN.IH(2) by auto
            moreover have "\<langle>M; \<A>; Sec\<rangle> \<turnstile>\<^sub>h\<^sub>o\<^sub>m t"
              using True DecomposeN.prems(1) DecomposeN.hyps(1) le_imp_less_Suc
                    Max_less_iff[of "insert l (steps ` K)" "Suc (Max (insert l (steps ` K)))"]
                    Ana_keys_finite[OF DecomposeN.hyps(2)]
              by blast
            ultimately show ?thesis
              by (metis intruder_deduct_hom.DecomposeH[OF _ DecomposeN.hyps(2) _ DecomposeN.hyps(4)])
          next
            case False
            obtain g S where gS: "t = Fun g S" using DecomposeN.hyps(2,4) by (cases t) moura+
            hence *: "Fun g S \<sqsubseteq> t" "\<not>?P (Fun g S)" using False by force+
            have "(\<exists>l'\<le>l. \<langle>M; l'\<rangle> \<turnstile>\<^sub>n Fun g S) \<and> (\<forall>t\<in>set S. \<exists>l'<l. \<langle>M; l'\<rangle> \<turnstile>\<^sub>n t)"
              using homogeneous_subterm[of _ \<A> Sec] GSMP\<^sub>s\<^sub>t_subterm[of _ "trms\<^sub>l\<^sub>s\<^sub>t \<A>"] M
                    deduct_normalize[of M ?P, OF _ DecomposeN.hyps(1) *]
              by meson
            hence "\<exists>l'<l. \<langle>M; l'\<rangle> \<turnstile>\<^sub>n t\<^sub>i"
              using gS DecomposeN.hyps(2,4) Ana_fun_subterm[of g S K T]
              by auto
            hence "\<exists>l'<Suc (Max (insert l (steps ` K))). \<langle>M; l'\<rangle> \<turnstile>\<^sub>n t\<^sub>i"
              using Max_less_iff[of "insert l (steps ` K)" "Suc (Max (insert l (steps ` K)))"]
                    Ana_keys_finite[OF DecomposeN.hyps(2)]
                    less_trans[of _ l "Suc (Max (insert l (steps ` K)))"]
              by blast
            thus ?thesis using DecomposeN.prems(1,2,3) by meson
          qed
        qed auto
      qed
    qed
  } thus ?thesis using deduct_num_if_deduct m(1) by metis
qed


subsection {* Lemmata: Deduction reduction of parallel composable constraints *}

lemma par_comp_hom_deduct:
  assumes \<A>: "par_comp \<A> Sec"
  and M: "\<forall>m \<in> M\<^sub>1 \<union> M\<^sub>2. homogeneous\<^sub>l\<^sub>s\<^sub>t m \<A> Sec" "\<forall>m \<in> M\<^sub>1 \<union> M\<^sub>2. m \<in> GSMP\<^sub>s\<^sub>t (trms\<^sub>l\<^sub>s\<^sub>t \<A>)"
         "M\<^sub>1 \<subseteq> GSMP\<^sub>s\<^sub>t (trms_left\<^sub>l\<^sub>s\<^sub>t \<A>)" "M\<^sub>2 \<subseteq> GSMP\<^sub>s\<^sub>t (trms_right\<^sub>l\<^sub>s\<^sub>t \<A>)"
         "Discl \<subseteq> M\<^sub>1" "Discl \<subseteq> M\<^sub>2"
         "Discl \<subseteq> Sec \<union> {m. {} \<turnstile>\<^sub>c m}"
  and Sec: "\<forall>s \<in> Sec - Discl. \<not>\<langle>M\<^sub>1; \<A>; Sec\<rangle> \<turnstile>\<^sub>h\<^sub>o\<^sub>m s"
           "\<forall>s \<in> Sec - Discl. \<not>\<langle>M\<^sub>2; \<A>; Sec\<rangle> \<turnstile>\<^sub>h\<^sub>o\<^sub>m s" 
  and t: "\<langle>M\<^sub>1 \<union> M\<^sub>2; \<A>; Sec\<rangle> \<turnstile>\<^sub>h\<^sub>o\<^sub>m t"
  shows "t \<notin> Sec - Discl" 
        "t \<in> GSMP\<^sub>s\<^sub>t (trms_left\<^sub>l\<^sub>s\<^sub>t  \<A>) \<Longrightarrow> \<langle>M\<^sub>1; \<A>; Sec\<rangle> \<turnstile>\<^sub>h\<^sub>o\<^sub>m t"
        "t \<in> GSMP\<^sub>s\<^sub>t (trms_right\<^sub>l\<^sub>s\<^sub>t \<A>) \<Longrightarrow> \<langle>M\<^sub>2; \<A>; Sec\<rangle> \<turnstile>\<^sub>h\<^sub>o\<^sub>m t"
using t
proof (induction t rule: intruder_deduct_hom_induct)
  case (AxiomH t)
  show t_not_Sec: "t \<notin> Sec - Discl"
  proof
    assume "t \<in> Sec - Discl"
    hence "\<not>\<langle>M\<^sub>1;\<A>;Sec\<rangle> \<turnstile>\<^sub>h\<^sub>o\<^sub>m t" "\<not>\<langle>M\<^sub>2;\<A>;Sec\<rangle> \<turnstile>\<^sub>h\<^sub>o\<^sub>m t" using Sec by auto
    thus False using AxiomH by auto
  qed
  have "t \<in> M\<^sub>1 \<Longrightarrow> t \<in> GSMP\<^sub>s\<^sub>t (trms_left\<^sub>l\<^sub>s\<^sub>t \<A>)"
       "t \<in> M\<^sub>2 \<Longrightarrow> t \<in> GSMP\<^sub>s\<^sub>t (trms_right\<^sub>l\<^sub>s\<^sub>t \<A>)"
    using M(3,4) AxiomH by auto
  moreover {
    assume "t \<in> M\<^sub>1 \<inter> M\<^sub>2"
    hence "t \<in> GSMP\<^sub>s\<^sub>t (trms_left\<^sub>l\<^sub>s\<^sub>t \<A>) \<inter> GSMP\<^sub>s\<^sub>t (trms_right\<^sub>l\<^sub>s\<^sub>t \<A>)"
      using M(3,4) AxiomH by auto
    hence "{} \<turnstile>\<^sub>c t \<or> t \<in> Discl" using \<A> t_not_Sec by auto
  } hence "t \<in> M\<^sub>1 \<inter> M\<^sub>2 \<Longrightarrow> {} \<turnstile>\<^sub>c t \<or> t \<in> Discl" by metis
  moreover have "t \<in> GSMP\<^sub>s\<^sub>t (trms_left\<^sub>l\<^sub>s\<^sub>t \<A>) \<inter> GSMP\<^sub>s\<^sub>t (trms_right\<^sub>l\<^sub>s\<^sub>t \<A>) \<Longrightarrow> {} \<turnstile>\<^sub>c t \<or> t \<in> Discl"
    using \<A> t_not_Sec by auto
  moreover have "homogeneous\<^sub>l\<^sub>s\<^sub>t t \<A> Sec" "t \<in> GSMP\<^sub>s\<^sub>t (trms\<^sub>l\<^sub>s\<^sub>t \<A>)" using AxiomH M(1,2) by auto
  moreover have "t \<in> Discl \<Longrightarrow> \<langle>M\<^sub>1; \<A>; Sec\<rangle> \<turnstile>\<^sub>h\<^sub>o\<^sub>m t" "t \<in> Discl \<Longrightarrow> \<langle>M\<^sub>2; \<A>; Sec\<rangle> \<turnstile>\<^sub>h\<^sub>o\<^sub>m t"
    using M(5,6) by auto
  ultimately show
       "t \<in> GSMP\<^sub>s\<^sub>t (trms_left\<^sub>l\<^sub>s\<^sub>t  \<A>) \<Longrightarrow> \<langle>M\<^sub>1; \<A>; Sec\<rangle> \<turnstile>\<^sub>h\<^sub>o\<^sub>m t"
       "t \<in> GSMP\<^sub>s\<^sub>t (trms_right\<^sub>l\<^sub>s\<^sub>t \<A>) \<Longrightarrow> \<langle>M\<^sub>2; \<A>; Sec\<rangle> \<turnstile>\<^sub>h\<^sub>o\<^sub>m t"
    using intruder_deduct_hom.AxiomH[of t M\<^sub>1 \<A> Sec]
          intruder_deduct_hom.AxiomH[of t M\<^sub>2 \<A> Sec]
          deduct_hom_if_synth[of t \<A> Sec "{}"]
          ideduct_hom_mono[of "{}" \<A> Sec t]
          AxiomH
    by fastforce+
next
  case (ComposeH T f)
  show "Fun f T \<in> GSMP\<^sub>s\<^sub>t (trms_left\<^sub>l\<^sub>s\<^sub>t \<A>) \<Longrightarrow> \<langle>M\<^sub>1; \<A>; Sec\<rangle> \<turnstile>\<^sub>h\<^sub>o\<^sub>m Fun f T"
  proof -
    assume "Fun f T \<in> GSMP\<^sub>s\<^sub>t (trms_left\<^sub>l\<^sub>s\<^sub>t \<A>)"
    hence "\<And>t. t \<in> set T \<Longrightarrow> t \<in> GSMP\<^sub>s\<^sub>t (trms_left\<^sub>l\<^sub>s\<^sub>t \<A>)"
      using GSMP\<^sub>s\<^sub>t_subterm[OF _ subtermeqI''] by auto
    thus ?thesis
      using ComposeH.IH(2) intruder_deduct_hom.ComposeH[OF ComposeH.hyps(1,2) _ ComposeH.hyps(4,5)]
      by simp
  qed
  moreover show "Fun f T \<in> GSMP\<^sub>s\<^sub>t (trms_right\<^sub>l\<^sub>s\<^sub>t \<A>) \<Longrightarrow> \<langle>M\<^sub>2; \<A>; Sec\<rangle> \<turnstile>\<^sub>h\<^sub>o\<^sub>m Fun f T"
  proof -
    assume "Fun f T \<in> GSMP\<^sub>s\<^sub>t (trms_right\<^sub>l\<^sub>s\<^sub>t \<A>)"
    hence "\<And>t. t \<in> set T \<Longrightarrow> t \<in> GSMP\<^sub>s\<^sub>t (trms_right\<^sub>l\<^sub>s\<^sub>t \<A>)"
      using GSMP\<^sub>s\<^sub>t_subterm[OF _ subtermeqI''] by auto
    thus ?thesis
      using ComposeH.IH(3) intruder_deduct_hom.ComposeH[OF ComposeH.hyps(1,2) _ ComposeH.hyps(4,5)]
      by simp
  qed
  ultimately show "Fun f T \<notin> Sec - Discl"
    using Sec ComposeH.hyps(5) SMP_append'[of "trms_left\<^sub>l\<^sub>s\<^sub>t \<A>" "trms_right\<^sub>l\<^sub>s\<^sub>t \<A>"]
          trms\<^sub>l\<^sub>s\<^sub>t_union[of \<A>]
    by auto
next
  case (DecomposeH t K T t\<^sub>i)
  have ti_subt: "t\<^sub>i \<sqsubseteq> t" using Ana_subterm[OF DecomposeH.hyps(2)] \<open>t\<^sub>i \<in> T\<close> by auto
  have t: "homogeneous\<^sub>l\<^sub>s\<^sub>t t \<A> Sec" "t \<in> GSMP\<^sub>s\<^sub>t (trms\<^sub>l\<^sub>s\<^sub>t \<A>)"
    using DecomposeH.hyps(1) hom_deduct_if_hom_ik M(1,2)
    by auto
  have ti: "homogeneous\<^sub>l\<^sub>s\<^sub>t t\<^sub>i \<A> Sec" "t\<^sub>i \<in> GSMP\<^sub>s\<^sub>t (trms\<^sub>l\<^sub>s\<^sub>t \<A>)"
    using intruder_deduct_hom.DecomposeH[OF DecomposeH.hyps] hom_deduct_if_hom_ik M(1,2) by auto
  { assume *: "t\<^sub>i \<in> GSMP\<^sub>s\<^sub>t (trms_left\<^sub>l\<^sub>s\<^sub>t \<A>)" "t \<in> GSMP\<^sub>s\<^sub>t (trms_left\<^sub>l\<^sub>s\<^sub>t \<A>)"
    hence "\<And>k. k \<in> K \<Longrightarrow> \<langle>M\<^sub>1;\<A>;Sec\<rangle> \<turnstile>\<^sub>h\<^sub>o\<^sub>m k"
      using GSMP\<^sub>s\<^sub>t_Ana_key[OF _ DecomposeH.hyps(2)] DecomposeH.IH(5) by auto
    hence "\<langle>M\<^sub>1;\<A>;Sec\<rangle> \<turnstile>\<^sub>h\<^sub>o\<^sub>m t\<^sub>i" "t\<^sub>i \<notin> Sec - Discl"
      using Sec DecomposeH.IH(2)[OF *(2)]
            intruder_deduct_hom.DecomposeH[OF _ DecomposeH.hyps(2) _ \<open>t\<^sub>i \<in> T\<close>]
      by force+
  } moreover {
    assume *: "t\<^sub>i \<in> GSMP\<^sub>s\<^sub>t (trms_left\<^sub>l\<^sub>s\<^sub>t \<A>)" "t \<in> GSMP\<^sub>s\<^sub>t (trms_right\<^sub>l\<^sub>s\<^sub>t \<A>)"
    hence "t\<^sub>i \<in> Sec \<union> {m. {} \<turnstile>\<^sub>c m}" using \<A> GSMP\<^sub>s\<^sub>t_subterm[OF *(2) ti_subt] by auto
    moreover have "\<And>k. k \<in> K \<Longrightarrow> \<langle>M\<^sub>2;\<A>;Sec\<rangle> \<turnstile>\<^sub>h\<^sub>o\<^sub>m k"
      using *(2) GSMP\<^sub>s\<^sub>t_Ana_key[OF _ DecomposeH.hyps(2)] DecomposeH.IH(6) by auto
    ultimately have "t\<^sub>i \<notin> Sec - Discl" "{} \<turnstile>\<^sub>c t\<^sub>i \<or> t\<^sub>i \<in> Discl"
      using Sec M(7) DecomposeH.IH(3)[OF *(2)]
            intruder_deduct_hom.DecomposeH[OF _ DecomposeH.hyps(2) _ \<open>t\<^sub>i \<in> T\<close>]
      by force+
    hence "\<langle>M\<^sub>1;\<A>;Sec\<rangle> \<turnstile>\<^sub>h\<^sub>o\<^sub>m t\<^sub>i" "\<langle>M\<^sub>2;\<A>;Sec\<rangle> \<turnstile>\<^sub>h\<^sub>o\<^sub>m t\<^sub>i" "t\<^sub>i \<notin> Sec - Discl"
      using M(5,6) deduct_hom_if_synth[THEN ideduct_hom_mono] ti by auto
   } moreover {
    assume *: "t\<^sub>i \<in> GSMP\<^sub>s\<^sub>t (trms_right\<^sub>l\<^sub>s\<^sub>t \<A>)" "t \<in> GSMP\<^sub>s\<^sub>t (trms_right\<^sub>l\<^sub>s\<^sub>t \<A>)"
    hence "\<And>k. k \<in> K \<Longrightarrow> \<langle>M\<^sub>2;\<A>;Sec\<rangle> \<turnstile>\<^sub>h\<^sub>o\<^sub>m k"
      using GSMP\<^sub>s\<^sub>t_Ana_key[OF _ DecomposeH.hyps(2)] DecomposeH.IH(6) by auto
    hence "\<langle>M\<^sub>2;\<A>;Sec\<rangle> \<turnstile>\<^sub>h\<^sub>o\<^sub>m t\<^sub>i" "t\<^sub>i \<notin> Sec - Discl"
      using DecomposeH.IH(3)[OF *(2)] Sec
            intruder_deduct_hom.DecomposeH[OF _ DecomposeH.hyps(2) _ \<open>t\<^sub>i \<in> T\<close>]
      by force+
  } moreover {
    assume *: "t\<^sub>i \<in> GSMP\<^sub>s\<^sub>t (trms_right\<^sub>l\<^sub>s\<^sub>t \<A>)" "t \<in> GSMP\<^sub>s\<^sub>t (trms_left\<^sub>l\<^sub>s\<^sub>t \<A>)"
    hence "t\<^sub>i \<in> Sec \<union> {m. {} \<turnstile>\<^sub>c m}" using \<A> GSMP\<^sub>s\<^sub>t_subterm[OF *(2) ti_subt] by auto
    moreover have "\<And>k. k \<in> K \<Longrightarrow> \<langle>M\<^sub>1;\<A>;Sec\<rangle> \<turnstile>\<^sub>h\<^sub>o\<^sub>m k"
      using *(2) GSMP\<^sub>s\<^sub>t_Ana_key[OF _ DecomposeH.hyps(2)] DecomposeH.IH(5) by auto
    ultimately have "t\<^sub>i \<notin> Sec - Discl" "{} \<turnstile>\<^sub>c t\<^sub>i \<or> t\<^sub>i \<in> Discl"
      using Sec M(7) DecomposeH.IH(2)[OF *(2)]
            intruder_deduct_hom.DecomposeH[OF _ DecomposeH.hyps(2) _ \<open>t\<^sub>i \<in> T\<close>]
      by force+
    hence "\<langle>M\<^sub>1;\<A>;Sec\<rangle> \<turnstile>\<^sub>h\<^sub>o\<^sub>m t\<^sub>i" "\<langle>M\<^sub>2;\<A>;Sec\<rangle> \<turnstile>\<^sub>h\<^sub>o\<^sub>m t\<^sub>i" "t\<^sub>i \<notin> Sec - Discl"
      using M(5,6) deduct_hom_if_synth[THEN ideduct_hom_mono] ti by auto
  } moreover have
        "t\<^sub>i \<in> GSMP\<^sub>s\<^sub>t (trms_left\<^sub>l\<^sub>s\<^sub>t \<A>) \<or> t\<^sub>i \<in> GSMP\<^sub>s\<^sub>t (trms_right\<^sub>l\<^sub>s\<^sub>t \<A>)"
        "t \<in> GSMP\<^sub>s\<^sub>t (trms_left\<^sub>l\<^sub>s\<^sub>t \<A>) \<or> t \<in> GSMP\<^sub>s\<^sub>t (trms_right\<^sub>l\<^sub>s\<^sub>t \<A>)"
      using t(2) ti(2) SMP_append'[of "trms_left\<^sub>l\<^sub>s\<^sub>t \<A>" "trms_right\<^sub>l\<^sub>s\<^sub>t \<A>"] trms\<^sub>l\<^sub>s\<^sub>t_union[of \<A>]
      by auto
  ultimately show
        "t\<^sub>i \<notin> Sec - Discl"
        "t\<^sub>i \<in> GSMP\<^sub>s\<^sub>t (trms_left\<^sub>l\<^sub>s\<^sub>t  \<A>) \<Longrightarrow> \<langle>M\<^sub>1; \<A>; Sec\<rangle> \<turnstile>\<^sub>h\<^sub>o\<^sub>m t\<^sub>i"
        "t\<^sub>i \<in> GSMP\<^sub>s\<^sub>t (trms_right\<^sub>l\<^sub>s\<^sub>t \<A>) \<Longrightarrow> \<langle>M\<^sub>2; \<A>; Sec\<rangle> \<turnstile>\<^sub>h\<^sub>o\<^sub>m t\<^sub>i"
      by auto
qed

lemma par_comp_deduct_left:
  assumes \<A>: "par_comp \<A> Sec"
  and M: "\<forall>m\<in>M\<^sub>1 \<union> M\<^sub>2. homogeneous\<^sub>l\<^sub>s\<^sub>t m \<A> Sec" "\<forall>m\<in>M\<^sub>1 \<union> M\<^sub>2. m \<in> GSMP\<^sub>s\<^sub>t (trms\<^sub>l\<^sub>s\<^sub>t \<A>)"
         "M\<^sub>1 \<subseteq> GSMP\<^sub>s\<^sub>t (trms_left\<^sub>l\<^sub>s\<^sub>t \<A>)" "M\<^sub>2 \<subseteq> GSMP\<^sub>s\<^sub>t (trms_right\<^sub>l\<^sub>s\<^sub>t \<A>)"
         "Discl \<subseteq> M\<^sub>1" "Discl \<subseteq> M\<^sub>2"
  and t: "M\<^sub>1 \<union> M\<^sub>2 \<turnstile> t" "t \<in> GSMP\<^sub>s\<^sub>t (trms_left\<^sub>l\<^sub>s\<^sub>t \<A>)"
  and Discl: "Discl \<subseteq> Sec \<union> {m. {} \<turnstile>\<^sub>c m}"
  shows "M\<^sub>1 \<turnstile> t \<or> (\<exists>s \<in> Sec - Discl. M\<^sub>1 \<turnstile> s \<or> M\<^sub>2 \<turnstile> s)"
using t
proof (induction t rule: intruder_deduct_induct)
  case (Axiom t) thus ?case
  proof (cases "t \<in> Sec - Discl \<or> {} \<turnstile>\<^sub>c t")
    case True thus ?thesis
      using Axiom ideduct_mono[of "{}" t]
      by (cases "t \<in> Sec - Discl") force+
  next
    case False
    hence "t \<notin> Sec - Discl" "\<not>{} \<turnstile>\<^sub>c t" "t \<in> GSMP\<^sub>s\<^sub>t (trms_left\<^sub>l\<^sub>s\<^sub>t \<A>)" using Axiom by auto
    hence "t \<notin> GSMP\<^sub>s\<^sub>t (trms_right\<^sub>l\<^sub>s\<^sub>t \<A>) \<or> t \<in> Discl" using \<A> unfolding GSMP_disjoint_def by auto
    hence "t \<notin> M\<^sub>2 \<or> t \<in> M\<^sub>1 \<or> {} \<turnstile>\<^sub>c t" using M by auto
    thus ?thesis using Axiom deduct_if_synth[THEN ideduct_mono] by auto
  qed
next
  case (Compose T f)
  hence "Fun f T \<in> GSMP\<^sub>s\<^sub>t (trms_left\<^sub>l\<^sub>s\<^sub>t \<A>)" using Compose.prems by auto
  hence IH: "\<And>t. t \<in> set T \<Longrightarrow> M\<^sub>1 \<turnstile> t \<or> (\<exists>s \<in> Sec - Discl. M\<^sub>1 \<turnstile> s \<or> M\<^sub>2 \<turnstile> s)"
    using Compose.IH by auto
  show ?case
  proof (cases "\<forall>t \<in> set T. M\<^sub>1 \<turnstile> t")
    case True thus ?thesis by (metis intruder_deduct.Compose[OF Compose.hyps(1,2)])
  qed (metis IH)
next
  case (Decompose t K T t\<^sub>i)
  have hom_ik: "\<forall>m\<in>M\<^sub>1 \<union> M\<^sub>2. homogeneous\<^sub>l\<^sub>s\<^sub>t m \<A> Sec \<and> m \<in> GSMP\<^sub>s\<^sub>t (trms\<^sub>l\<^sub>s\<^sub>t \<A>)" using M by auto

  have hom_GSMP_ti: "homogeneous\<^sub>l\<^sub>s\<^sub>t t\<^sub>i \<A> Sec" "t\<^sub>i \<in> GSMP\<^sub>s\<^sub>t (trms\<^sub>l\<^sub>s\<^sub>t \<A>)"
    using Decompose.prems(1) GSMP_is_homogeneous \<A>
          trms\<^sub>l\<^sub>s\<^sub>t_union[of \<A>] SMP_append'[of "trms_left\<^sub>l\<^sub>s\<^sub>t \<A>" "trms_right\<^sub>l\<^sub>s\<^sub>t \<A>"]
    unfolding par_comp_def
    by auto

  have "\<langle>M\<^sub>1 \<union> M\<^sub>2; \<A>; Sec\<rangle> \<turnstile>\<^sub>h\<^sub>o\<^sub>m t\<^sub>i"
    using intruder_deduct.Decompose[OF Decompose.hyps]
          hom_deduct_if_deduct[OF \<A> hom_ik _ hom_GSMP_ti]
    by metis
  hence "\<langle>M\<^sub>1; \<A>; Sec\<rangle> \<turnstile>\<^sub>h\<^sub>o\<^sub>m t\<^sub>i \<or> (\<exists>s\<in>Sec-Discl. \<langle>M\<^sub>1;\<A>;Sec\<rangle> \<turnstile>\<^sub>h\<^sub>o\<^sub>m s \<or> \<langle>M\<^sub>2;\<A>;Sec\<rangle> \<turnstile>\<^sub>h\<^sub>o\<^sub>m s)"
    using par_comp_hom_deduct(2)[OF \<A> M Discl _ _ _ Decompose.prems(1)]
    by auto
  thus ?case using deduct_if_hom_deduct[of _ \<A> Sec] by auto
qed

lemma par_comp_deduct_right:
  assumes \<A>: "par_comp \<A> Sec"
  and M: "\<forall>m\<in>M\<^sub>1 \<union> M\<^sub>2. homogeneous\<^sub>l\<^sub>s\<^sub>t m \<A> Sec" "\<forall>m\<in>M\<^sub>1 \<union> M\<^sub>2. m \<in> GSMP\<^sub>s\<^sub>t (trms\<^sub>l\<^sub>s\<^sub>t \<A>)"
         "M\<^sub>1 \<subseteq> GSMP\<^sub>s\<^sub>t (trms_left\<^sub>l\<^sub>s\<^sub>t \<A>)" "M\<^sub>2 \<subseteq> GSMP\<^sub>s\<^sub>t (trms_right\<^sub>l\<^sub>s\<^sub>t \<A>)"
         "Discl \<subseteq> M\<^sub>1" "Discl \<subseteq> M\<^sub>2"
  and t: "M\<^sub>1 \<union> M\<^sub>2 \<turnstile> t" "t \<in> GSMP\<^sub>s\<^sub>t (trms_right\<^sub>l\<^sub>s\<^sub>t \<A>)"
  and Discl: "Discl \<subseteq> Sec \<union> {m. {} \<turnstile>\<^sub>c m}"
  shows "M\<^sub>2 \<turnstile> t \<or> (\<exists>s \<in> Sec - Discl. M\<^sub>1 \<turnstile> s \<or> M\<^sub>2 \<turnstile> s)"
using t
proof (induction t rule: intruder_deduct_induct)
  case (Axiom t) thus ?case
  proof (cases "t \<in> Sec - Discl \<or> {} \<turnstile>\<^sub>c t")
    case True thus ?thesis
      using Axiom ideduct_mono[of "{}" t]
      by (cases "t \<in> Sec - Discl") force+
  next
    case False
    hence "t \<notin> Sec - Discl" "\<not>{} \<turnstile>\<^sub>c t" "t \<in> GSMP\<^sub>s\<^sub>t (trms_right\<^sub>l\<^sub>s\<^sub>t \<A>)" using Axiom by auto
    hence "t \<notin> GSMP\<^sub>s\<^sub>t (trms_left\<^sub>l\<^sub>s\<^sub>t \<A>) \<or> t \<in> Discl" using \<A> unfolding GSMP_disjoint_def by auto
    hence "t \<in> M\<^sub>2 \<or> t \<notin> M\<^sub>1 \<or> {} \<turnstile>\<^sub>c t" using M by auto
    thus ?thesis using Axiom deduct_if_synth[THEN ideduct_mono] by auto
  qed
next
  case (Compose T f)
  hence "Fun f T \<in> GSMP\<^sub>s\<^sub>t (trms_right\<^sub>l\<^sub>s\<^sub>t \<A>)" using Compose.prems by auto
  hence IH: "\<And>t. t \<in> set T \<Longrightarrow> M\<^sub>2 \<turnstile> t \<or> (\<exists>s \<in> Sec - Discl. M\<^sub>1 \<turnstile> s \<or> M\<^sub>2 \<turnstile> s)"
    using Compose.IH by auto
  show ?case
  proof (cases "\<forall>t \<in> set T. M\<^sub>2 \<turnstile> t")
    case True thus ?thesis by (metis intruder_deduct.Compose[OF Compose.hyps(1,2)])
  qed (metis IH)
next
  case (Decompose t K T t\<^sub>i)
  have hom_ik: "\<forall>m\<in>M\<^sub>1 \<union> M\<^sub>2. homogeneous\<^sub>l\<^sub>s\<^sub>t m \<A> Sec \<and> m \<in> GSMP\<^sub>s\<^sub>t (trms\<^sub>l\<^sub>s\<^sub>t \<A>)" using M by auto

  have hom_GSMP_ti: "homogeneous\<^sub>l\<^sub>s\<^sub>t t\<^sub>i \<A> Sec" "t\<^sub>i \<in> GSMP\<^sub>s\<^sub>t (trms\<^sub>l\<^sub>s\<^sub>t \<A>)"
    using Decompose.prems(1) GSMP_is_homogeneous \<A>
          trms\<^sub>l\<^sub>s\<^sub>t_union[of \<A>] SMP_append'[of "trms_left\<^sub>l\<^sub>s\<^sub>t \<A>" "trms_right\<^sub>l\<^sub>s\<^sub>t \<A>"]
    unfolding par_comp_def
    by auto

  have "\<langle>M\<^sub>1 \<union> M\<^sub>2; \<A>; Sec\<rangle> \<turnstile>\<^sub>h\<^sub>o\<^sub>m t\<^sub>i"
    using intruder_deduct.Decompose[OF Decompose.hyps]
          hom_deduct_if_deduct[OF \<A> hom_ik _ hom_GSMP_ti]
    by metis
  hence "\<langle>M\<^sub>2; \<A>; Sec\<rangle> \<turnstile>\<^sub>h\<^sub>o\<^sub>m t\<^sub>i \<or> (\<exists>s\<in>Sec-Discl. \<langle>M\<^sub>1;\<A>;Sec\<rangle> \<turnstile>\<^sub>h\<^sub>o\<^sub>m s \<or> \<langle>M\<^sub>2;\<A>;Sec\<rangle> \<turnstile>\<^sub>h\<^sub>o\<^sub>m s)"
    using par_comp_hom_deduct(3)[OF \<A> M Discl _ _ _ Decompose.prems(1)]
    by auto
  thus ?case using deduct_if_hom_deduct[of _ \<A> Sec] by auto
qed


subsection {* Theorem: Parallel compositionality for labeled constraints *}
lemma par_comp_prefix: assumes "par_comp (A@B) M" shows "par_comp A M"
proof -
  have GSMP: "GSMP_disjoint (trms_left\<^sub>l\<^sub>s\<^sub>t A \<union> trms_left\<^sub>l\<^sub>s\<^sub>t B) (trms_right\<^sub>l\<^sub>s\<^sub>t A \<union> trms_right\<^sub>l\<^sub>s\<^sub>t B) M"
    and Bsnd: "\<forall>t. (\<star>, snd\<langle>t\<rangle>) \<notin> set (A@B)"
    using assms lst_left_right_append unlabel_append trms\<^sub>s\<^sub>t_append
    unfolding par_comp_def by metis+
  hence "\<forall>t. (\<star>, snd\<langle>t\<rangle>) \<notin> set A" "GSMP_disjoint (trms_left\<^sub>l\<^sub>s\<^sub>t A) (trms_right\<^sub>l\<^sub>s\<^sub>t A) M"
    using GSMP(1) SMP_append' by auto
  thus ?thesis using assms unfolding par_comp_def by blast
qed

theorem par_comp_constr_typed:
  assumes \<A>: "par_comp \<A> Sec"
  and \<I>: "\<I> \<Turnstile> \<langle>unlabel \<A>\<rangle>" "interpretation\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<I>" "wt\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<I>" "\<forall>x. wf\<^sub>t\<^sub>r\<^sub>m (\<I> x)"
  shows "(\<I> \<Turnstile> \<langle>left_unl \<A>\<rangle> \<and> \<I> \<Turnstile> \<langle>right_unl \<A>\<rangle>) \<or> (\<exists>\<A>'. prefixeq \<A>' \<A> \<and> \<A>' leaks Sec under \<I>)"
proof -
  let ?L = "\<lambda>\<A>'. \<exists>t \<in> Sec - declassified\<^sub>l\<^sub>s\<^sub>t \<A>' \<I>. \<lbrakk>{}; left_unl \<A>':snd\<langle>t\<rangle>\<rbrakk>\<^sub>d \<I> \<or> \<lbrakk>{}; right_unl \<A>':snd\<langle>t\<rangle>\<rbrakk>\<^sub>d \<I>"
  have "\<lbrakk>{}; unlabel \<A>\<rbrakk>\<^sub>d \<I>" using \<I> by simp
  with \<A> have "\<lbrakk>{}; left_unl \<A>\<rbrakk>\<^sub>d \<I> \<and> \<lbrakk>{}; right_unl \<A>\<rbrakk>\<^sub>d \<I> \<or> (\<exists>\<A>'. prefixeq \<A>' \<A> \<and> ?L \<A>')"
  proof (induction "unlabel \<A>" arbitrary: \<A> rule: List.rev_induct)
    case Nil
    hence "\<A> = []" using unlabel_nil_only_if_nil by simp
    thus ?case by auto
  next
    case (snoc b B \<A>)
    hence disj: "GSMP_disjoint (trms_left\<^sub>l\<^sub>s\<^sub>t \<A>) (trms_right\<^sub>l\<^sub>s\<^sub>t \<A>) Sec" by auto

    obtain a A where a: "\<A> = A@[a]" "a = (l1, b) \<or> a = (l2, b) \<or> a = (\<star>, b)"
      using unlabel_snoc_inv[OF snoc.hyps(2)[symmetric]] by moura
    hence A: "\<A> = A@[(l1, b)] \<or> \<A> = A@[(l2, b)] \<or> \<A> = A@[(\<star>, b)]" by metis

    have 1: "B = unlabel A" using a snoc.hyps(2) unlabel_append[of A "[a]"] by auto
    have 2: "par_comp A Sec" using par_comp_prefix snoc.prems(1) a by metis
    have 3: "\<lbrakk>{}; unlabel A\<rbrakk>\<^sub>d \<I>" using a 1 snoc.prems(2) snoc.hyps(2) strand_sem_split_left_d by metis
    have IH: "\<lbrakk>{}; left_unl A\<rbrakk>\<^sub>d \<I> \<and> \<lbrakk>{}; right_unl A\<rbrakk>\<^sub>d \<I> \<or> (\<exists>\<A>'. prefixeq \<A>' A \<and> ?L \<A>')"
      by (rule snoc.hyps(1)[OF 1 2 3])

    show ?case
    proof (cases "\<lbrakk>{}; left_unl A\<rbrakk>\<^sub>d \<I> \<and> \<lbrakk>{}; right_unl A\<rbrakk>\<^sub>d \<I>")
      case False
      then obtain \<A>' where \<A>': "prefixeq \<A>' A" "?L \<A>'" by (metis IH)
      hence "prefixeq \<A>' (A@[a])" using a prefixeq_append(1)[of _ A "[a]"] by simp
      thus ?thesis using \<A>'(2) a by auto
    next
      case True
      hence IH': "\<lbrakk>{}; left_unl A\<rbrakk>\<^sub>d \<I>" "\<lbrakk>{}; right_unl A\<rbrakk>\<^sub>d \<I>" by metis+
      show ?thesis
      proof (cases b)
        case (Send t)
        hence "ik\<^sub>s\<^sub>t (unlabel A) \<cdot>set \<I> \<turnstile> t \<cdot> \<I>"
          using a \<open>\<lbrakk>{}; unlabel \<A>\<rbrakk>\<^sub>d \<I>\<close> strand_sem_split_right[of "{}" "unlabel A" "unlabel [a]" \<I>]
                unlabel_append[of A "[a]"]
          by auto
        hence *: "(ik\<^sub>s\<^sub>t (left_unl A) \<cdot>set \<I>) \<union> (ik\<^sub>s\<^sub>t (right_unl A) \<cdot>set \<I>) \<turnstile> t \<cdot> \<I>"
          using left_right_ik_union_is_unlabel_ik subst_all_union by metis

        have "ik\<^sub>s\<^sub>t (left_unl \<A>) = ik\<^sub>s\<^sub>t (left_unl A)" "ik\<^sub>s\<^sub>t (right_unl \<A>) = ik\<^sub>s\<^sub>t (right_unl A)"
          using Send A lst_left_right_append[of A] by auto
        hence **:
            "ik\<^sub>s\<^sub>t (left_unl A) \<cdot>set \<I> \<subseteq> GSMP\<^sub>s\<^sub>t (trms_left\<^sub>l\<^sub>s\<^sub>t \<A>)"
            "ik\<^sub>s\<^sub>t (right_unl A) \<cdot>set \<I> \<subseteq> GSMP\<^sub>s\<^sub>t (trms_right\<^sub>l\<^sub>s\<^sub>t \<A>)"
          using ik_right_subst_GSMP_subset[OF \<I>(3,4,2), of \<A>]
                ik_left_subst_GSMP_subset[OF \<I>(3,4,2), of \<A>]
          by auto

        note Discl =
          declassified_left_ik_subset[of A \<I>] declassified_right_ik_subset[of A \<I>]
          declassified_left_GSMP_subset[OF \<I>(3,4,2), of A]
          declassified_right_GSMP_subset[OF \<I>(3,4,2), of A]
          declassified_secrets_subset[OF disj \<I>(3,4,2)] declassified_append[of A "[a]" \<I>]

        have Sec: "\<forall>s \<in> Sec. \<forall>s' \<in> subterms s. {} \<turnstile>\<^sub>c s' \<or> s' \<in> Sec" "ground Sec"
          using \<A> by auto

        have "\<forall>m\<in>(ik\<^sub>s\<^sub>t (left_unl \<A>) \<cdot>set \<I>) \<union> (ik\<^sub>s\<^sub>t (right_unl \<A>) \<cdot>set \<I>). homogeneous\<^sub>l\<^sub>s\<^sub>t m \<A> Sec"
             "\<forall>m\<in>(ik\<^sub>s\<^sub>t (left_unl \<A>) \<cdot>set \<I>) \<union> (ik\<^sub>s\<^sub>t (right_unl \<A>) \<cdot>set \<I>). m \<in> GSMP\<^sub>s\<^sub>t (trms\<^sub>l\<^sub>s\<^sub>t \<A>)"
             "ik\<^sub>s\<^sub>t (left_unl \<A>) \<cdot>set \<I> \<subseteq> GSMP\<^sub>s\<^sub>t (trms_left\<^sub>l\<^sub>s\<^sub>t \<A>)"
             "ik\<^sub>s\<^sub>t (right_unl \<A>) \<cdot>set \<I> \<subseteq> GSMP\<^sub>s\<^sub>t (trms_right\<^sub>l\<^sub>s\<^sub>t \<A>)"
          using secrets_are_homogeneous[OF Sec(1)]
                GSMP_left_is_homogeneous[OF disj]
                GSMP_right_is_homogeneous[OF disj]
                ik_left_subst_GSMP_subset[OF \<I>(3,4,2), of \<A>]
                ik_right_subst_GSMP_subset[OF \<I>(3,4,2), of \<A>]
          apply blast
          using ik_left_subst_GSMP_subset[OF \<I>(3,4,2), of \<A>]
                ik_right_subst_GSMP_subset[OF \<I>(3,4,2), of \<A>]
                GSMP_union[of \<A>]
          by auto
        hence M: "\<forall>m\<in>(ik\<^sub>s\<^sub>t (left_unl A) \<cdot>set \<I>) \<union> (ik\<^sub>s\<^sub>t (right_unl A) \<cdot>set \<I>). homogeneous\<^sub>l\<^sub>s\<^sub>t m \<A> Sec"
                 "\<forall>m\<in>(ik\<^sub>s\<^sub>t (left_unl A) \<cdot>set \<I>) \<union> (ik\<^sub>s\<^sub>t (right_unl A) \<cdot>set \<I>). m \<in> GSMP\<^sub>s\<^sub>t (trms\<^sub>l\<^sub>s\<^sub>t \<A>)"
                 "ik\<^sub>s\<^sub>t (left_unl A) \<cdot>set \<I> \<subseteq> GSMP\<^sub>s\<^sub>t (trms_left\<^sub>l\<^sub>s\<^sub>t \<A>)"
                 "ik\<^sub>s\<^sub>t (right_unl A) \<cdot>set \<I> \<subseteq> GSMP\<^sub>s\<^sub>t (trms_right\<^sub>l\<^sub>s\<^sub>t \<A>)"
          using a(1) by auto

        have "a = (l1, b) \<or> a = (l2, b)" using Send a snoc.prems(1) by auto
        hence "(a = (l1, snd\<langle>t\<rangle>) \<and> t \<cdot> \<I> \<in> GSMP\<^sub>s\<^sub>t (trms_left\<^sub>l\<^sub>s\<^sub>t \<A>)) \<or>
               (a = (l2, snd\<langle>t\<rangle>) \<and> t \<cdot> \<I> \<in> GSMP\<^sub>s\<^sub>t (trms_right\<^sub>l\<^sub>s\<^sub>t \<A>))"
          using Send a(1) trms_left\<^sub>l\<^sub>s\<^sub>t_append[of A "[a]"] trms_right\<^sub>l\<^sub>s\<^sub>t_append[of A "[a]"]
                GSMP_wt_substI[OF _ \<I>(3,4,2)]
          by force
        hence
            "a = (l1, snd\<langle>t\<rangle>) \<or> a = (l2, snd\<langle>t\<rangle>)"
            "a = (l1, snd\<langle>t\<rangle>) \<Longrightarrow> \<lbrakk>{}; right_unl (A@[a])\<rbrakk>\<^sub>d \<I>"
            "a = (l1, snd\<langle>t\<rangle>) \<Longrightarrow> ik\<^sub>s\<^sub>t (left_unl A) \<cdot>set \<I> \<turnstile> t \<cdot> \<I> \<Longrightarrow> \<lbrakk>{}; left_unl (A@[a])\<rbrakk>\<^sub>d \<I>"
            "a = (l1, snd\<langle>t\<rangle>) \<Longrightarrow> t \<cdot> \<I> \<in> GSMP\<^sub>s\<^sub>t (trms_left\<^sub>l\<^sub>s\<^sub>t \<A>)"
            "a = (l2, snd\<langle>t\<rangle>) \<Longrightarrow> \<lbrakk>{}; left_unl (A@[a])\<rbrakk>\<^sub>d \<I>"
            "a = (l2, snd\<langle>t\<rangle>) \<Longrightarrow> ik\<^sub>s\<^sub>t (right_unl A) \<cdot>set \<I> \<turnstile> t \<cdot> \<I> \<Longrightarrow> \<lbrakk>{}; right_unl (A@[a])\<rbrakk>\<^sub>d \<I>"
            "a = (l2, snd\<langle>t\<rangle>) \<Longrightarrow> t \<cdot> \<I> \<in> GSMP\<^sub>s\<^sub>t (trms_right\<^sub>l\<^sub>s\<^sub>t \<A>)"
          using IH' lst_left_right_append[of A]
          by auto
        hence
            "(\<lbrakk>{}; left_unl \<A>\<rbrakk>\<^sub>d \<I> \<and> \<lbrakk>{}; right_unl \<A>\<rbrakk>\<^sub>d \<I>) \<or>
             (\<exists>s \<in> Sec-declassified\<^sub>l\<^sub>s\<^sub>t A \<I>. ik\<^sub>s\<^sub>t (left_unl A) \<cdot>set \<I> \<turnstile> s \<or> ik\<^sub>s\<^sub>t (right_unl A) \<cdot>set \<I> \<turnstile> s)"
          using par_comp_deduct_left[OF snoc.prems(1) M Discl(1,2) *]
                par_comp_deduct_right[OF snoc.prems(1) M Discl(1,2) *]
                A a Discl(3,4,5,6)
          by auto
        moreover have "prefixeq A \<A>" using A by auto
        moreover have "\<And>s. s \<in> Sec \<Longrightarrow> s \<cdot> \<I> = s" using Sec(2) by auto
        hence "\<forall>s \<in> Sec - declassified\<^sub>l\<^sub>s\<^sub>t A \<I>. ik\<^sub>s\<^sub>t (left_unl A)  \<cdot>set \<I> \<turnstile> s \<longrightarrow> \<lbrakk>{}; left_unl A:snd\<langle>s\<rangle>\<rbrakk>\<^sub>d \<I>"
              "\<forall>s \<in> Sec - declassified\<^sub>l\<^sub>s\<^sub>t A \<I>. ik\<^sub>s\<^sub>t (right_unl A) \<cdot>set \<I> \<turnstile> s \<longrightarrow> \<lbrakk>{}; right_unl A:snd\<langle>s\<rangle>\<rbrakk>\<^sub>d \<I>"
          using strand_sem_append_d[OF IH'(1)] strand_sem_append_d[OF IH'(2)]
          by auto
        ultimately show ?thesis by metis
      next
        case (Receive t) thus ?thesis using IH' a lst_left_right_append[of A "[a]"] by force
      next
        case (Equality t t')
        hence *: "\<And>M. \<lbrakk>M; [t \<doteq> t']\<rbrakk>\<^sub>d \<I>"
          using a \<open>\<lbrakk>{}; unlabel \<A>\<rbrakk>\<^sub>d \<I>\<close> strand_sem_split_right[of "{}" "unlabel A" "unlabel [a]" \<I>]
                unlabel_append[of A "[a]"]
          by auto
        show ?thesis
          using a lst_left_right_append[of A "[a]"] Equality
                strand_sem_append_d[OF IH'(1) *] strand_sem_append_d[OF IH'(2) *]
          by force
      next
        case (Inequality X t t')
        hence *: "\<And>M. \<lbrakk>M; [\<nexists>X\<langle>t \<doteq> t'\<rangle>]\<rbrakk>\<^sub>d \<I>"
          using a \<open>\<lbrakk>{}; unlabel \<A>\<rbrakk>\<^sub>d \<I>\<close> strand_sem_split_right[of "{}" "unlabel A" "unlabel [a]" \<I>]
                unlabel_append[of A "[a]"]
          by auto
        show ?thesis
          using a lst_left_right_append[of A "[a]"] Inequality
                strand_sem_append_d[OF IH'(1) *] strand_sem_append_d[OF IH'(2) *]
          by force
      qed
    qed
  qed
  thus ?thesis using \<I>(1) unfolding strand_leaks\<^sub>l\<^sub>s\<^sub>t_def by simp
qed

theorem par_comp_constr:
  assumes \<A>: "par_comp \<A> Sec" "typing_cond (unlabel \<A>)"
  and \<I>: "\<I> \<Turnstile> \<langle>unlabel \<A>\<rangle>" "interpretation\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<I>"
  shows "\<exists>\<I>\<^sub>\<tau>. interpretation\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<I>\<^sub>\<tau> \<and> wt\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<I>\<^sub>\<tau> \<and> (\<forall>x. wf\<^sub>t\<^sub>r\<^sub>m (\<I>\<^sub>\<tau> x)) \<and>
              ((\<I>\<^sub>\<tau> \<Turnstile> \<langle>left_unl \<A>\<rangle> \<and> \<I>\<^sub>\<tau> \<Turnstile> \<langle>right_unl \<A>\<rangle>) \<or>
              (\<exists>\<A>'. prefixeq \<A>' \<A> \<and> \<A>' leaks Sec under \<I>\<^sub>\<tau>))"
proof -
  from \<A>(2) have *:
    "wf\<^sub>s\<^sub>t {} (unlabel \<A>)"
    "FV\<^sub>s\<^sub>t (unlabel \<A>) \<inter> bvars\<^sub>s\<^sub>t (unlabel \<A>) = {}"
    "tfr\<^sub>s\<^sub>t (trms\<^sub>l\<^sub>s\<^sub>t \<A>)" "eqs_tfr\<^sub>s\<^sub>t (unlabel \<A>)"
    "Ana_invar_subst (ik\<^sub>s\<^sub>t (unlabel \<A>) \<union> eqs_rhs\<^sub>s\<^sub>t (unlabel \<A>))"
   unfolding typing_cond_def by metis+
  obtain \<I>\<^sub>\<tau> where \<I>\<^sub>\<tau>: "\<I>\<^sub>\<tau> \<Turnstile> \<langle>unlabel \<A>\<rangle>" "interpretation\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<I>\<^sub>\<tau>" "wt\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<I>\<^sub>\<tau>" "\<forall>x. wf\<^sub>t\<^sub>r\<^sub>m (\<I>\<^sub>\<tau> x)"
    using wt_attack_if_tfr_attack_d[OF * \<I>(2,1)] by metis
  show ?thesis using par_comp_constr_typed[OF \<A>(1) \<I>\<^sub>\<tau>] \<I>\<^sub>\<tau>(2,3,4) by auto
qed


subsection {* Definitions: Labeled protocols *}
type_synonym ('a,'b) labeledprotocol = "('a,'b) labeledstrand set"

abbreviation wf\<^sub>l\<^sub>s\<^sub>t where
  "wf\<^sub>l\<^sub>s\<^sub>t V \<A> \<equiv> wf\<^sub>s\<^sub>t V (unlabel \<A>)"

definition wf\<^sub>l\<^sub>s\<^sub>t\<^sub>s::"('fun,'var) labeledprotocol \<Rightarrow> bool" where
  "wf\<^sub>l\<^sub>s\<^sub>t\<^sub>s \<S> \<equiv> (\<forall>S \<in> \<S>. wf\<^sub>l\<^sub>s\<^sub>t {} S) \<and> (\<forall>S \<in> \<S>. \<forall>S' \<in> \<S>. FV\<^sub>l\<^sub>s\<^sub>t S \<inter> bvars\<^sub>l\<^sub>s\<^sub>t S' = {})"

definition wf\<^sub>l\<^sub>s\<^sub>t\<^sub>s'::"('fun,'var) labeledprotocol \<Rightarrow> ('fun,'var) labeledstrand \<Rightarrow> bool" where
  "wf\<^sub>l\<^sub>s\<^sub>t\<^sub>s' \<S> \<A> \<equiv> (\<forall>S \<in> \<S>. wf\<^sub>s\<^sub>t (vars2\<^sub>l\<^sub>s\<^sub>t \<A>) (unlabel S)) \<and>
                 (\<forall>S \<in> \<S>. \<forall>S' \<in> \<S>. FV\<^sub>l\<^sub>s\<^sub>t S \<inter> bvars\<^sub>l\<^sub>s\<^sub>t S' = {}) \<and>
                 (\<forall>S \<in> \<S>. FV\<^sub>l\<^sub>s\<^sub>t S \<inter> bvars\<^sub>l\<^sub>s\<^sub>t \<A> = {}) \<and>
                 (\<forall>S \<in> \<S>. FV\<^sub>l\<^sub>s\<^sub>t \<A> \<inter> bvars\<^sub>l\<^sub>s\<^sub>t S = {})"

definition typing_cond_prot where
  "typing_cond_prot \<P> \<equiv>
    wf\<^sub>l\<^sub>s\<^sub>t\<^sub>s \<P> \<and> tfr\<^sub>s\<^sub>t (\<Union>(trms\<^sub>l\<^sub>s\<^sub>t ` \<P>)) \<and> (\<forall>S \<in> \<P>. eqs_tfr\<^sub>s\<^sub>t (unlabel S)) \<and>
    Ana_invar_subst (\<Union>(ik\<^sub>s\<^sub>t ` unlabel ` \<P>) \<union> \<Union>(eqs_rhs\<^sub>s\<^sub>t ` unlabel ` \<P>))"

definition par_comp_prot where
  "par_comp_prot \<P> Sec \<equiv>
    (GSMP_disjoint (\<Union>\<A> \<in> \<P>. trms_left\<^sub>l\<^sub>s\<^sub>t \<A>) (\<Union>\<A> \<in> \<P>. trms_right\<^sub>l\<^sub>s\<^sub>t \<A>) Sec) \<and>
    (\<forall>\<A> \<in> \<P>. \<forall>t. (\<star>, snd\<langle>t\<rangle>) \<notin> set \<A>) \<and> ground Sec \<and>
    (\<forall>s \<in> Sec. \<forall>s' \<in> subterms s. {} \<turnstile>\<^sub>c s' \<or> s' \<in> Sec) \<and>
    typing_cond_prot \<P>"


subsection {* Lemmata: Labeled protocols *}
lemma wf\<^sub>l\<^sub>s\<^sub>t\<^sub>s_eqs_wf\<^sub>l\<^sub>s\<^sub>t\<^sub>s'[simp]: "wf\<^sub>l\<^sub>s\<^sub>t\<^sub>s S = wf\<^sub>l\<^sub>s\<^sub>t\<^sub>s' S []"
unfolding wf\<^sub>l\<^sub>s\<^sub>t\<^sub>s_def wf\<^sub>l\<^sub>s\<^sub>t\<^sub>s'_def unlabel_def by auto

lemma par_comp_prot_impl_par_comp:
  assumes "par_comp_prot \<P> Sec" "\<A> \<in> \<P>"
  shows "par_comp \<A> Sec"
proof -
  have *: "GSMP_disjoint (\<Union>\<A> \<in> \<P>. trms_left\<^sub>l\<^sub>s\<^sub>t \<A>) (\<Union>\<A> \<in> \<P>. trms_right\<^sub>l\<^sub>s\<^sub>t \<A>) Sec"
    using assms(1) unfolding par_comp_prot_def by metis
  have "GSMP_disjoint (trms_left\<^sub>l\<^sub>s\<^sub>t \<A>) (trms_right\<^sub>l\<^sub>s\<^sub>t \<A>) Sec"
    using GSMP_disjoint_subset[OF *] assms(2) by auto
  thus ?thesis using assms unfolding par_comp_prot_def par_comp_def by metis
qed

lemma typing_cond_prot_impl_typing_cond:
  assumes "typing_cond_prot \<P>" "\<A> \<in> \<P>"
  shows "typing_cond (unlabel \<A>)"
proof -
  have 1: "wf\<^sub>s\<^sub>t {} (unlabel \<A>)" "FV\<^sub>l\<^sub>s\<^sub>t \<A> \<inter> bvars\<^sub>l\<^sub>s\<^sub>t \<A> = {}"
    using assms unfolding typing_cond_prot_def wf\<^sub>l\<^sub>s\<^sub>t\<^sub>s_def by auto

  have "tfr\<^sub>s\<^sub>t (\<Union>(trms\<^sub>l\<^sub>s\<^sub>t ` \<P>))"
       "trms\<^sub>l\<^sub>s\<^sub>t \<A> \<subseteq> \<Union>(trms\<^sub>l\<^sub>s\<^sub>t ` \<P>)"
       "SMP\<^sub>s\<^sub>t (trms\<^sub>l\<^sub>s\<^sub>t \<A>) - Var`\<V> \<subseteq> SMP\<^sub>s\<^sub>t (\<Union>(trms\<^sub>l\<^sub>s\<^sub>t ` \<P>)) - Var`\<V>"
    using assms SMP_mono[of "trms\<^sub>l\<^sub>s\<^sub>t \<A>" "\<Union>(trms\<^sub>l\<^sub>s\<^sub>t ` \<P>)"]
    unfolding typing_cond_prot_def
    by (metis, auto)
  hence 2: "tfr\<^sub>s\<^sub>t (trms\<^sub>l\<^sub>s\<^sub>t \<A>)" by (meson subsetD)

  have 3: "eqs_tfr\<^sub>s\<^sub>t (unlabel \<A>)" using assms unfolding typing_cond_prot_def by auto

  have "\<Union>(subterms ` (ik\<^sub>s\<^sub>t (unlabel \<A>) \<union> eqs_rhs\<^sub>s\<^sub>t (unlabel \<A>))) \<subseteq>
        \<Union>(subterms ` (\<Union>(ik\<^sub>s\<^sub>t ` unlabel ` \<P>) \<union> \<Union>(eqs_rhs\<^sub>s\<^sub>t ` unlabel ` \<P>)))"
    using assms(2) by auto
  hence 4: "Ana_invar_subst (ik\<^sub>s\<^sub>t (unlabel \<A>) \<union> eqs_rhs\<^sub>s\<^sub>t (unlabel \<A>))"
    using assms SMP_mono unfolding typing_cond_prot_def Ana_invar_subst by (meson subsetD)
  show ?thesis using 1 2 3 4 unfolding typing_cond_def by blast
qed



subsection {* Theorem: Parallel compositionality for labeled protocols *}
definition left_prot where
  "left_prot P \<equiv> (\<forall>l \<in> P. \<forall>s \<in> set l. \<not>is_Label2 s)"

definition right_prot where
  "right_prot P \<equiv> (\<forall>l \<in> P. \<forall>s \<in> set l. \<not>is_Label1 s)"

definition composed_prot where
  "composed_prot \<P>\<^sub>1 \<P>\<^sub>2 \<equiv> {\<A>. left \<A> \<in> \<P>\<^sub>1 \<and> right \<A> \<in> \<P>\<^sub>2}"

definition left_secure_prot where
  "left_secure_prot P Sec attack \<equiv> (\<forall>\<A> \<in> P. suffixeq [(l1, snd\<langle>Fun attack []\<rangle>)] \<A> \<longrightarrow> 
     (\<forall>\<I>\<^sub>\<tau>. (interpretation\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<I>\<^sub>\<tau> \<and> wt\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<I>\<^sub>\<tau> \<and> (\<forall>x. wf\<^sub>t\<^sub>r\<^sub>m (\<I>\<^sub>\<tau> x))) \<longrightarrow>
            \<not>(\<I>\<^sub>\<tau> \<Turnstile> \<langle>left_unl \<A>\<rangle>) \<and>
            (\<forall>\<A>'. prefixeq \<A>' \<A> \<longrightarrow> (\<forall>t \<in> Sec-declassified\<^sub>l\<^sub>s\<^sub>t \<A>' \<I>\<^sub>\<tau>. \<not>(\<I>\<^sub>\<tau> \<Turnstile> \<langle>left_unl \<A>':snd\<langle>t\<rangle>\<rangle>)))))"

definition right_leaks where
  "right_leaks \<A> Sec \<equiv> (\<exists>\<A>' \<I>\<^sub>\<tau>. interpretation\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<I>\<^sub>\<tau> \<and> wt\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<I>\<^sub>\<tau> \<and> (\<forall>x. wf\<^sub>t\<^sub>r\<^sub>m (\<I>\<^sub>\<tau> x)) \<and>
      prefixeq \<A>' \<A> \<and> (\<exists>t \<in> Sec - declassified\<^sub>l\<^sub>s\<^sub>t \<A>' \<I>\<^sub>\<tau>. (\<I>\<^sub>\<tau> \<Turnstile> \<langle>right_unl \<A>':snd\<langle>t\<rangle>\<rangle>)))"

definition unsat where
  "unsat \<A> \<equiv> (\<forall>\<I>. interpretation\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<I> \<longrightarrow> \<not>\<I> \<Turnstile> \<langle>unlabel \<A>\<rangle>)"

theorem par_comp_constr_prot:
  assumes P: "P = composed_prot P1 P2" "par_comp_prot P Sec" "left_prot P1"
  and left_secure: "left_secure_prot P1 Sec attack"
  shows "\<forall>\<A> \<in> P. suffixeq [(l1, snd\<langle>Fun attack []\<rangle>)] \<A> \<longrightarrow> unsat \<A> \<or> right_leaks \<A> Sec"
proof -
  { fix \<A> \<A>' assume \<A>: "\<A> = \<A>':(l1, snd\<langle>Fun attack []\<rangle>)" "\<A> \<in> P"
    let ?P = "\<exists>\<A>' \<I>\<^sub>\<tau>. interpretation\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<I>\<^sub>\<tau> \<and> wt\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<I>\<^sub>\<tau> \<and> (\<forall>x. wf\<^sub>t\<^sub>r\<^sub>m (\<I>\<^sub>\<tau> x)) \<and> prefixeq \<A>' \<A> \<and>
                       (\<exists>t \<in> Sec - declassified\<^sub>l\<^sub>s\<^sub>t \<A>' \<I>\<^sub>\<tau>. (\<I>\<^sub>\<tau> \<Turnstile> \<langle>right_unl \<A>':snd\<langle>t\<rangle>\<rangle>))"
    have tcp: "typing_cond_prot P" using P(2) unfolding par_comp_prot_def by simp
    have par_comp: "par_comp \<A> Sec" "typing_cond (unlabel \<A>)"
      using par_comp_prot_impl_par_comp[OF P(2) \<A>(2)]
            typing_cond_prot_impl_typing_cond[OF tcp \<A>(2)]
      by metis+
  
    have "unlabel (left \<A>) = left_unl \<A>" "left_unl \<A> = left_unl (left \<A>)"
         "\<And>A. A \<in> P1 \<Longrightarrow> left A = A" "left \<A> \<in> P1"
         "left \<A> = (left \<A>'):(l1, snd\<langle>Fun attack []\<rangle>)"
      using P(1,3) \<A>
      by (auto simp add: left_def unlabel_def left_prot_def composed_prot_def)
    moreover {
      fix A assume "prefixeq A \<A>"
      hence "\<exists>B. prefixeq B (left \<A>) \<and> left_unl A = left_unl B \<and>
                 (\<forall>I. declassified\<^sub>l\<^sub>s\<^sub>t A I = declassified\<^sub>l\<^sub>s\<^sub>t B I)"
        unfolding prefixeq_def left_def declassified\<^sub>l\<^sub>s\<^sub>t_def
        by force
    }
    ultimately have *: "\<forall>\<I>\<^sub>\<tau>. interpretation\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<I>\<^sub>\<tau> \<and> wt\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<I>\<^sub>\<tau> \<and> (\<forall>x. wf\<^sub>t\<^sub>r\<^sub>m (\<I>\<^sub>\<tau> x)) \<longrightarrow>
                             \<not>\<I>\<^sub>\<tau> \<Turnstile> \<langle>left_unl \<A>\<rangle> \<and> (\<forall>\<A>'. prefixeq \<A>' \<A> \<longrightarrow>
                                  (\<forall>t\<in>Sec - declassified\<^sub>l\<^sub>s\<^sub>t \<A>' \<I>\<^sub>\<tau>. \<not>\<I>\<^sub>\<tau> \<Turnstile> \<langle>left_unl \<A>':snd\<langle>t\<rangle>\<rangle>))"
      using left_secure unfolding left_secure_prot_def composed_prot_def suffixeq_def by metis
    { fix \<I> assume \<I>: "interpretation\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<I>" "\<I> \<Turnstile> \<langle>unlabel \<A>\<rangle>"
      obtain \<I>\<^sub>\<tau> where \<I>\<^sub>\<tau>:
          "interpretation\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<I>\<^sub>\<tau>" "wt\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<I>\<^sub>\<tau>" "\<forall>x. wf\<^sub>t\<^sub>r\<^sub>m (\<I>\<^sub>\<tau> x)"
          "\<exists>\<A>'. prefixeq \<A>' \<A> \<and> \<A>' leaks Sec under \<I>\<^sub>\<tau>"
        using par_comp_constr[OF par_comp \<I>(2,1)] * by moura
      hence "\<exists>\<A>'. prefixeq \<A>' \<A> \<and> (\<exists>t \<in> Sec - declassified\<^sub>l\<^sub>s\<^sub>t \<A>' \<I>\<^sub>\<tau>. \<I>\<^sub>\<tau> \<Turnstile> \<langle>right_unl \<A>':snd\<langle>t\<rangle>\<rangle>)"
        using \<I>\<^sub>\<tau>(4) * unfolding strand_leaks\<^sub>l\<^sub>s\<^sub>t_def by auto
      hence ?P using \<I>\<^sub>\<tau>(1,2,3) by auto
    } hence "unsat \<A> \<or> right_leaks \<A> Sec" by (metis unsat_def right_leaks_def)
  } thus ?thesis unfolding suffixeq_def by metis
qed

end

end
